package AOSICFG;
use utf8;
use LWP::UserAgent;
use Digest::MD5 qw(md5 md5_base64);
use Digest::SHA qw(sha1_base64);
use XML::Simple;
use Sys::Hostname;
use MIME::Base64;
use Net::LDAP;
use Net::LDAP::Control::PasswordPolicy;
use Net::LDAP::Constant qw( LDAP_CONTROL_PASSWORDPOLICY  );
use Encode;

write_ws_log('===== Configuration '.$config::aosi_version.' =====');
write_ws_log('-=- Debug level ' . $config::aosi_debug_level . ' -=-');

write_ws_log('-=- Getting Debian version -=-');
if(-e '/etc/debian_version'){
    $debian_version = _get_pwd('/etc/debian_version');
}
$dpkg = '';
if($debian_version =~ /^centos/i){  # Hosting and USS
    if(-e $config::lib_dir.'/dpkg.pl'){
        $dpkg = $config::lib_dir.'/dpkg.pl';
    }
}
else {
    $ENV{'PATH'} = '/usr/bin:/bin';
    $dpkg = 'dpkg -l \'*-aai\' |grep ^ii';
}
write_ws_log('-=- Getting Debian package versions -=-');
$out  = qx($dpkg);
@out  = split(/[\n\r]/, $out);
AOSICFG::write_ws_log('-=- run='.$dpkg.' -=-') if $config::aosi_debug_level > 3;
foreach $line (@out) {
    ($status, $name, $version, $desc) = split(/\s+/, $line);
    $package_version{$name} = $version;
    $name = XML::Simple::escape_value($dummy, $name);
    $version = XML::Simple::escape_value($dummy, $version);
}
write_ws_log('-=- '.dump_hash(\%package_version).' -=-');

write_ws_log('-=- Reading LDAP password from ' . $config::ldap_pwd_file . ' -=-');
$ldap_pwd = _get_pwd($config::ldap_pwd_file);

@ldap_admins = get_list($config::ldap_admins_file);
write_ws_log('-=- Reading admin user list from ' . $config::ldap_admins_file . ': '.join(", ", @ldap_admins).' -=-');
# Backward compatibilty
@valid_users = @ldap_admins;

write_ws_log('-=- Getting remote schema file from ' . $config::schema_url . ' -=-');
my ($schema_ok, $schema_file) = get_url_cached($config::schema_url, $config::remote_schema_file, $config::local_schema_file);
if (-e $config::local_schema_file) {
    get_bin_attribs($config::local_schema_file, $config::bin_attrib_file);
    get_req_attribs($config::local_schema_file, $config::req_attrib_file);
}

write_ws_log('-=- Getting enumerated values for attributes -=-');
foreach $attrib (keys %$config::attribute_xml) {
    my ($att_code, $att_file, $att_xml) = get_url_cached(
        $config::schema_base . $config::attribute_xml->{$attrib},
        $config::varlib_dir . '/remote_' . $config::attribute_xml->{$attrib},
        $config::varlib_dir . '/' . $config::attribute_xml->{$attrib}
    );
    if ($att_code == 0) {
        $attr_restriction->{$attrib} = get_attrib_list($att_xml);
        write_ws_log('-=- ' . join(", ", @{ $attr_restriction->{$attrib} }) . ' -=-');
    }
}

write_ws_log('-=- Use IP list for direct access: ' . $config::direct_access . ' -=-');
# Backward compatibilty
$config::use_valid_ip = $config::direct_access;

if($config::direct_access == 1){
    @direct_access_ip = get_list($config::direct_access_ip_file);
    write_ws_log('-=- Reading direct access IP list from ' . $config::direct_access_ip_file . ': '.join(", ", @direct_access_ip).' -=-');
    # Backward compatibilty
    @valid_ip = @direct_access_ip;
}

write_ws_log('-=- Reading valid IP list for monitoring from ' . $config::valid_mon_file . ' -=-');
@valid_mon = get_list($config::valid_mon_file);

write_ws_log('-=- Reading binary attribute list from ' . $config::bin_attrib_file . ' -=-');
@bin_attribs = get_list($config::bin_attrib_file);
@bin_attribs = map { chomp; $_; } @bin_attribs;

write_ws_log('-=- Reading required attribute list from ' . $config::req_attrib_file . ' -=-');
@req_attribs = get_list($config::req_attrib_file);
@req_attribs = map { chomp; $_; } @req_attribs;
write_ws_log('-=- Configuration done -=-');

sub get_bin_attribs {
    my ($file_in, $file_out) = @_;
    my ($xml, $att, $type, @bin, $att_name);

    $xml = XMLin($file_in, KeyAttr => []);
    open(BIN, ">$config::bin_attrib_file") or die "$0: error opening $config::bin_attrib_file for writting ($!)\n";
    foreach $att (sort { $a->{'ldapname'} cmp $b->{'ldapname'} } @{ $xml->{'attribute'} }) {

        #print "att: ".$att->{'ldapname'}.";type: ";
        @keys = keys %{ $att->{'type'} };
        $type = $keys[0];
        if ($type =~ /binary/i) {
            $att_name = $att->{'ldapname'};
            push(@bin, $att_name);
            print BIN $att_name, "\n";
        }
    }
    close(BIN);
    write_ws_log('-=- Writting binary attribute list (' . join(",", @bin) . ') to ' . $config::bin_attrib_file . ' -=-');
}

sub get_req_attribs {
    my ($file_in, $file_out) = @_;
    my ($att, $att_name, $type, $xml, @bin, @keys);

    $xml = XMLin($file_in, KeyAttr => []);
    open(MUST, ">$config::req_attrib_file") or die "$0: error opening $config::req_attrib_file for writting ($!)\n";
    foreach $att (sort { $a->{'ldapname'} cmp $b->{'ldapname'} } @{ $xml->{'attribute'} }) {
        $must = $att->{'status'}->{'must'};
        if ($must =~ /true/i) {
            $att_name = $att->{'ldapname'};
            push(@req, $att_name);
            print MUST $att_name . "\n";
        }
    }
    close(MUST);
    write_ws_log('-=- Writting required attribute list (' . join(",", @req) . ') to ' . $config::req_attrib_file . ' -=-');
}

sub get_xml {
    my ($url) = @_;
    my ($ua, $response);

    $ua = LWP::UserAgent->new;
    $ua->timeout(10);
    $ua->env_proxy;

    $response = $ua->get($url);

    if ($response->is_success) {
        return 0, $response->content;
    }
    else {
        return $response->code;
    }
}

sub get_url_cached {
    my ($url, $remote_file, $local_file) = @_;
    my ($code, $content, $code_out, $remote_md5, $local_md5, $file_out, @lines);

    ($code, $content) = get_xml($url);
    write_ws_log('-=- Get from url ' . $url . ' to ' . $remote_file . ' and cache to ' . $local_file . ': code=' . $code . ' -=-');

    $code_out = 0;
    $file_out = $local_file;
    if ($code == 0) {
        open(XML, ">$remote_file") or die "($!)\n";
        print XML $content;
        close(XML);

        $remote_md5 = get_file_md5($remote_file);
        $local_md5  = get_file_md5($local_file);
        write_ws_log('-=- Remote file MD5:  ' . $remote_md5 . ' -=-');
        write_ws_log('-=- Local file MD5:   ' . $local_md5 . ' -=-');

        if ($local_md5 ne $remote_md5) {
            write_ws_log('-=- Files are different:  ' . $remote_file . ' => ' . $local_file . ' -=-');
            if (-e $local_file) {
                rename($local_file, $local_file . '.bkp');
            }
            rename($remote_file, $local_file);
        }
        else {
            write_ws_log('-=- Files are equal. No update needed! -=-');
        }
    }
    else {
        write_ws_log('-=- Using local file ' . $local_file . ' -=-');
        if (-e $local_file) {
            open(XML, "$local_file") or die "$0: error opening $local_file for reading ($!)\n";
            @lines = <XML>;
            close(XML);
            $content = join("", @lines);
        }
        else {
            write_ws_log('-=- PANIC!!!! Local file does not exist! -=-');
            $code_out = -1;
            $file_out = '';
        }
    }

    return $code_out, $file_out, $content;
}

sub get_file_md5 {
    my ($file) = @_;
    my ($md5, $out);

    $out = '';
    if (-e $file) {
        open(XML, "$file") or die "$0: error opening $file for reading ($!)\n";
        $md5 = Digest::MD5->new;
        $md5->addfile(*XML);
        $out = $md5->b64digest;
        close(XML);
    }

    return $out;
}

sub get_attrib_list {
    my ($xml_in) = @_;
    my ($att, @out);

    $xml = XMLin($xml_in, KeyAttr => [], ForceArray => ['vocabularyentry'], NormalizeSpace => 2);
    foreach $att (@{ $xml->{'vocabularyentry'} }) {
        push(@out, $att->{'key'});
    }
    return \@out;
}

sub timestamp {
    my ($sec, $min, $hour, $mday, $mon, $year, $wday) = localtime(time);
    my (@mon_name);

    @mon_name = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
    return sprintf("%s %02d %02d:%02d:%02d", $mon_name[$mon], $mday, $hour, $min, $sec);
}

sub hr_datetime {
    my ($sec, $min, $hour, $mday, $mon, $year, $wday) = localtime(time);
    my (@mon_name);

    return sprintf("<d>%02d.%02d.%04d.</d><t>%02d:%02d:%02d</t>", $mday, $mon + 1, 1900 + $year, $hour, $min, $sec);
}

sub write_ws_log {
    my ($stg) = @_;
    open(LOG, ">>$config::ws_log_file") or die "$0: can not open file $config::ws_log_file for appending ($!)\n";
    $stg =~ s/\n//gm;

    $stg =~ s|<attribute ldapname="userPassword">\s*<value>[^<]+</value>\s*</attribute>|<attribute ldapname="userPassword"><value>*****</value></attribute>|g;
    print LOG timestamp(), " ", hostname(), " aosi[", $$, "]: ", utfize($stg), "\n";
    close(LOG);
}

sub write_action_log {
    my ($user, $action, $dn, $code, $attrs) = @_;
    open(ALOG, ">>$config::action_log_file") or die "$0: can not open file $config::action_log_file for appending ($!)\n";
    print ALOG "<e>", hr_datetime(), "<w>$user</w><a>$action</a><s>$dn</s><c>$code</c><p>$attrs</p></e>\n";
    close(ALOG);
}

sub write_search_log {
    my ($user, $action, $base, $filter, $attrs, $code) = @_;

    if($config::write_search_log == 1){
        open(SLOG, ">>$config::search_log_file") or die "$0: can not open file $config::search_log_file for appending ($!)\n";
        print SLOG "<e>", hr_datetime(), "<w>$user</w><a>$action</a><b>$base</b><f>$filter</f><c>$code</c><p>$attrs</p><i>".$AOSI::client_ip."</i></e>\n";
        close(SLOG);
    }
}


sub write_pid {
    my ($file) = @_;

    open(FILE, ">$file") or die "$0: can not open file $file for writting ($!)\n";
    print FILE $$;
    close(FILE);
}

sub _get_pwd {
    my ($file) = @_;
    my ($real);

    open(FILE, "$file") or die "$0: can not open file $file for reading ($!)\n";
    $real = <FILE>;
    close(FILE);
    chomp($real);
    return $real;
}

sub get_pwd {
    my ($file) = @_;
    if ($file eq $config::ldap_pwd_file) {
        return _get_pwd($file);
    }
    else {
        return '';
    }
}

sub get_list {
    my ($file) = @_;
    my (@real, $i);

    open(FILE, "$file") or die "$0: can not open file $file for reading ($!)\n";
    @real = <FILE>;
    close(FILE);

    chomp(@real);
    return @real;
}

sub user_split {
    my ($in) = @_;
    my ($uid, $realm);

    $uid = $in;
    if ($uid =~ /\@/) {
        ($uid, $realm) = split(/\@/, $uid);
    }
    return $uid, $realm;
}

sub _user_is_admin {
    my ($in) = @_;
    my ($ok, $realm, $uid, $user, @users);

    $ok = 0;
    ($uid, $realm) = user_split($in);
    foreach $user (@ldap_admins) {
        chomp($user);
        if ($user eq $uid) {
            $ok++;
        }
    }
    write_ws_log('| User ' . $in . ' is valid: ' . $ok) if $config::aosi_debug_level > 2;
    return $ok;
}

sub user_is_admin {
    my ($in) = @_;
    return _user_is_admin($in);
}

# Backward compatibilty
sub user_is_valid {
    my ($in) = @_;
    return _user_is_admin($in);
}

sub _ip_has_direct_access {
    my ($in) = @_;
    my ($ip, $ok);

    $ok = 0;
    foreach $ip (@direct_access_ip) {
        chomp($ip);
        if ($ip eq $in) {
            $ok++;
        }
    }
    write_ws_log('| IP ' . $in . ' is valid: ' . $ok) if $config::aosi_debug_level > 2;
    return $ok;
}

sub ip_has_direct_access {
    my ($in) = @_;
    return _ip_has_direct_access($in);
}

# Backward compatibilty
sub ip_is_valid {
    my ($in) = @_;
    return _ip_has_direct_access($in);
}

sub _mon_is_valid {
    my ($in) = @_;
    my ($ip, $ok);

    $ok = 0;
    foreach $ip (@valid_mon) {
        chomp($ip);
        if ($ip eq $in) {
            $ok++;
        }
    }
    write_ws_log('| Mon IP ' . $in . ' is valid: ' . $ok) if $config::aosi_debug_level > 2;
    return $ok;
}

sub mon_is_valid {
    my ($in) = @_;
    return _mon_is_valid($in);
}

sub _is_bin {
    my ($attrib) = @_;
    my ($attr, $ok);

    $ok = 0;
    foreach $attr (@bin_attribs) {
        chomp($attr);
        if ($attr eq $attrib) {
            $ok++;
        }
    }
    write_ws_log('| Attrib ' . $attrib . ' is binary: ' . $ok) if $config::aosi_debug_level > 2;
    return $ok;
}

sub is_bin {
    my ($attrib) = @_;
    return _is_bin($attrib);
}

sub is_restriction_ok {
    my ($attrib, $value) = @_;
    my (@ok_vals, $val, $ok, $out, $msg);

    $out = 1;
    $msg = '';
    write_ws_log('| is_restriction_ok(in): att=' . $attrib . ';val=' . $value) if ($config::aosi_debug_level > 3 and $attrib ne 'userPassword');
    if($attrib eq 'hrEduPersonUniqueNumber'){
        if(($value =~ /^JMBG:\s*(\d+)/) and ($config::check_jmbg == 1)){
            if(!ck_jmbg($1)){
                $out = 0;
                $msg = 'JMBG '.$1.' is not valid!';
            }
        }
        if(($value =~ /^OIB:\s*(\d+)/) and ($config::check_oib == 1)){
            if(!ck_oib($1)){
                $out = 0;
                $msg = 'OIB '.$1.' is not valid!';
            }
        }
    }
    elsif(($attrib eq 'hrEduPersonOIB') or ($attrib eq 'hrEduOrgOIB')){
         if(!ck_oib($value) and ($value ne 'NONE')){
            $out = 0;
            $msg = 'OIB '.$value.' is not valid!';
         }
    }
    else {
        if (defined $attr_restriction->{$attrib}) {
            @ok_vals = @{ $attr_restriction->{$attrib} };
            $ok      = 0;
            foreach $val (@ok_vals) {
                if ($val eq $value) {
                    $ok++;
                }
                write_ws_log('| is_restriction_ok(cmp): ' . $val . ' ? ' . $value . ' = '. $ok) if $config::aosi_debug_level > 3;
            }
            $out = $ok;
            $msg =
                  'Values for attribute '
                . $attrib
                . ' are restricted to values found in '
                . $config::schema_base
                . $config::attribute_xml->{$attrib} . '!'
        }
    }
    write_ws_log('| is_restriction_ok(out): ' . $out) if $config::aosi_debug_level > 3;
    return $out, $msg;
}

sub dump_hash {
    my ($hash, $dont_dump) = @_;
    my ($out, $key, $val, $len);
    $out = '';

    foreach $key (keys %$hash) {
        $val = $hash->{$key};
        Encode::_utf8_on($val);
        if (ref($val) eq 'ARRAY') {
            $val = '[' . join(',', @$val) . ']';
        }
        $len = length($val);
        if ($len > $config::log_large_value_limit) {
            $val = '***Large value. Not dumped!***';
        }
        if ($key eq 'userPassword') {
            $val = '***';
        }
        $out .= $key . " => " . $val . ";";

    }
    return $out;
}

sub dump_array {
    my ($array, $dont_dump) = @_;
    my ($out, $i, $val, $len);
    $out = '';

    foreach $i (0 .. $#{$array}) {
        next if (($dont_dump eq 'defined') and ($dont_dump == $i));
        $val = $array->[$i];
        if (ref($val) eq 'ARRAY') {
            $val = '[' . join(',', @$val) . ']';
        }
        $len = length($val);
        if ($len > $config::log_large_value_limit) {
            $val = '***Large value. Not dumped!***';
        }
        $out .= $val . ";";

    }
    return $out;
}

sub utfize {
    my ($in) = @_;

    if (utf8::is_utf8($in)) {
        if (is_within_ascii($in)) {
            utf8::encode($in);
        }
    }
    return $in;
}

sub is_within_ascii {
    my $string = shift;

    # look for anything that isn't ascii or pass
    if ($string =~ /([^\x{00}-\x{7f}])/) {
        return 1;
    }
    else {
        return 0;
    }
}

######################################################################
# Intialize LDAP connection
######################################################################
sub ldap_init {
    my ($hostname, $port) = @_;
    my ($ldap, $ldap_error);

    $error = '';
    $ldap  = Net::LDAP->new(
        $hostname,
        port    => $port,
        debug   => $config::ldap_debug_level,
        timeout => $config::ldap_timeout
    );
    unless ($ldap) {
        AOSICFG::write_ws_log("| Connection to LDAP server $hostname:$port failed: $@");
        $error = "Connection to LDAP server $hostname:$port failed: $@";
    }
    return $ldap, $error;
}

######################################################################
#
######################################################################
sub xml2entries {
    my ($xml_in) = @_;
    my ($attr, $attrs, $ent, $hash, $i, $key, $val, $xml, %attr_exists, @entries, @keys, @vals);

    if (utf8::is_utf8($xml_in)) {
        utf8::encode($xml_in);
    }
    AOSICFG::write_ws_log("|/ XML2Entries: xml_in=$xml_in") if $config::aosi_debug_level > 3;
    $xml = XMLin($xml_in, ForceArray => [ 'entry', 'value', 'attribute' ], NormalizeSpace => 2);
    @entries = ();
    foreach $ent (@{ $xml->{'entry'} }) {
        $hash         = {};
        $attrs        = undef;
        $hash->{'dn'} = $ent->{'dn'};
        AOSICFG::write_ws_log("|| XML2Entries: dn = " . $ent->{'dn'} . "") if $config::aosi_debug_level > 3;
        $i = 0;
        foreach $attr (@{ $ent->{'attribute'} }) {
            $key = $attr->{'ldapname'};
            $val = $attr->{'value'};

            # If the <value> is empty it is represented as 'value' => [ {} ]
            if (ref($val) eq 'ARRAY') {
                @vals = @$val;
                if ($#vals == 0 and ref($vals[0]) eq 'HASH') {
                    @keys = keys %{ $vals[0] };
                    if ($#keys < 0) {
                        $val = [];
                    }
                }
            }
            AOSICFG::write_ws_log("|| XML2Entries: value = (" . join(", ",          @$val) . ")") if ($config::aosi_debug_level > 3 and $key ne 'userPassword');
            AOSICFG::write_ws_log("|| XML2Entries: att = $key; val = (" . join(",", @$val) . ")") if ($config::aosi_debug_level > 3 and $key ne 'userPassword');
            push(@$attrs, { 'att' => $key, 'val' => $val, 'id' => $i });
            $i++;
            $attr_exists{$key}++;
        }
        $hash->{'attrs'} = $attrs;
        AOSICFG::write_ws_log("|| XML2Entries(hash):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 3;
        push(@entries, $hash);
    }
    AOSICFG::write_ws_log("|\\ XML2Entries End") if $config::aosi_debug_level > 3;
    return \@entries, \%attr_exists;
}

######################################################################
# Separate hostname and port part
######################################################################
sub parse_server {
    my ($server) = @_;
    my ($hostname, $port);

    if (index($server, ':') > -1) {
        ($hostname, $port) = split(/:/, $server);
    }
    else {
        $hostname = $server;
        $port     = 389;
    }
    return $hostname, $port;
}

######################################################################
#
######################################################################
sub get_dn {
    my ($user, $base) = @_;
    my ($uid, $realm, $dn);

    ($uid, $realm) = split(/\@/, $user);
    if ($uid =~ /=/g) {
        $dn = $uid . ',' . $base;
    }
    else {
        $dn = 'uid=' . $uid . ',' . $base;
    }

    return $dn;
}

######################################################################
#
######################################################################
sub get_uid_base {
    my ($dn) = @_;
    my ($uid, $realm, $dummy, $base, @base);

    AOSICFG::write_ws_log("|/ get_uid_base(in):dn:$dn;") if $config::aosi_debug_level > 3;
    ($uid,   @base) = split(/,/, $dn);
    ($dummy, $uid)  = split(/=/, $uid);
    $base = join(",", @base);
    AOSICFG::write_ws_log("|\\ get_uid_base(out):uid:$uid;base:$base;") if $config::aosi_debug_level > 3;
    return $uid, $base;
}

######################################################################
#
######################################################################
sub ldapGenSearch {
    my ($service, $user, $pwd, $base, $filter, $attributes, $bind_user, $from, $size, $get_bin, $md5_in) = @_;
    my ($auth_code, $auth_error, $code,    $dec_pwd, $entry,        $hash,      $hostname,   $dummy);
    my ($i,         $is_admin,   $ldap,    $log,     $md5,          $mesg2,     $out,        $port, $show_bin, $to, $val, $attr, $cur_attr);
    my (%options,   @attrs,      @entries, @values,  $admin_filter, $my_filter, $ldap_error, $dn);
    my ($plugin_hash, $plugin_msg, $plugin_code, $plugin_xml, $temp, $pp_mesg, $ppolicy);

    $hash     = undef;
    $out      = '';
    $show_bin = $get_bin || 0;
    $log      = 'ldapGenSearch';

    AOSICFG::write_ws_log("| $log(in):user:$user;base:$base;filter:$filter;attrib:$attributes;user:$bind_user;md5:$md5_in;") if $config::aosi_debug_level > 2;

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = ldap_init($hostname, $port);
    $dn = get_dn($user, $base);
    $admin_filter = '';
    if ($bind_user == 1) {
        ($auth_code, $auth_error, $is_admin, $admin_filter, $pp_mesg) = AOSICFG::ck_auth($ldap, get_dn($user, $base), $pwd, 0, $ldap_error);
    }
    elsif ($bind_user == -1) {

        # Do nothing
        AOSICFG::write_ws_log("| $log:anonymous access");
        unless ($ldap) {
            $auth_code  = -18;
            $auth_error = $ldap_error;
        }
    }
    else {
        ($auth_code, $auth_error, $is_admin, $admin_filter, $pp_mesg) = AOSICFG::ck_auth($ldap, get_dn($user, $base), $pwd, 1, $ldap_error);
        Encode::_utf8_on($admin_filter);
        if ($is_admin == 0 and $ldap) {
            if ($auth_code == 0) {
                $auth_code  = -15;
                $auth_error = "User $user does not have admin privileges!";
            }
            if ($from > -1) {
                $hash->{'from'}  = 0;
                $hash->{'size'}  = 0;
                $hash->{'count'} = 0;
            }
        }
    }
    if ($auth_code == 0) {
		$ppolicy = Net::LDAP::Control::PasswordPolicy->new;
        if ($bind_user == 1) {
            $dec_pwd = decode_base64($pwd);
            $mesg2 = $ldap->bind(get_dn($user, $base), password => $dec_pwd, version => 3, control => [ $ppolicy ]);
            AOSICFG::write_ws_log("| $log(bind_as_user code): " . $mesg2->code) if $config::aosi_debug_level > 2;

        }
        elsif ($bind_user == -1) {
            $mesg2 = $ldap->bind(version => 3, control => [ $ppolicy ]);
            AOSICFG::write_ws_log("| $log(bind_as_anonymous code): " . $mesg2->code) if $config::aosi_debug_level > 2;

        }
        else {
            $mesg2 = $ldap->bind($config::ldap_user, password => AOSICFG::get_pwd($config::ldap_pwd_file), version => 3, control => [ $ppolicy ]);
            AOSICFG::write_ws_log("| $log(bind_as_admin code): " . $mesg2->code) if $config::aosi_debug_level > 2;
            if ($mesg2->code != 0) {
                AOSICFG::write_ws_log("| !!!$log: invalid admin password in " . $config::ldap_pwd_file);
            }
        }
        if ($filter =~ /^\(.*\)$/) {
            $my_filter = '(&' . $filter . '' . $admin_filter . ')';
        }
        else {
            $my_filter = '(&(' . $filter . ')' . $admin_filter . ')';
        }
        AOSICFG::write_ws_log("| $log(my_filter): " . $my_filter) if $config::aosi_debug_level > 2;
        %options = (
            base   => $base,
            filter => $my_filter,

            #sizelimit => 0,
            timelimit => 180
        );
        if ($attributes) {
            @attrs = split(/[,; ]/, $attributes);
            $options{'attrs'} = \@attrs;
        }
		AOSICFG::write_ws_log("| $log(search options): " . AOSICFG::dump_hash(\%options)) if $config::aosi_debug_level > 3;
        $mesg2          = $ldap->search(%options);
        $code           = $mesg2->code;
        $hash->{'code'} = $code;
        $to             = $from + $size - 1;
        $i              = -1;
        if ($code == 0) {
            @entries = $mesg2->entries;
          LOOP:
            foreach $entry (sort bydn @entries) {
                if ($from > -1) {
                    $i++;
                    if ($i < $from) {
                        next LOOP;
                    }
                    if ($i > $to) {
                        last;
                    }
                }
                $dn = $entry->dn();
                $out .= "<entry";
                if ($from > -1) {
                    $out .= " num=\"$i\"";
                }
                $out .= " dn=\"$dn\">";

				if($pp_mesg){
					$out .= "<attribute ldapname=\"pwdPolicySubEntry\">";
					$out .= "<value>";
					$out .= $pp_mesg;
					$out .= "</value>";                        
					$out .= "</attribute>";                        
				}
                foreach $attr ($entry->attributes) {
                    $cur_attr = $attr eq 'homePhone' ? 'homeTelephoneNumber' : $attr;
                    AOSICFG::write_ws_log("| $log(attr): $attr") if $config::aosi_debug_level > 3;
                    #$hash->{$attr} = join(', ', @{$entry->get_value( $attr, asref => 1 )});
                    $out .= "<attribute ldapname=\"" . $cur_attr . "\">";
                    @values = @{ $entry->get_value($attr, asref => 1) };
                    if (AOSICFG::is_bin($attr)) {
                        if ($md5_in eq '') {
                            foreach $val (@values) {
                                if ($show_bin == 1) {
                                    $out .= "<value>" . encode_base64($val) . "</value>";
                                }
                                else {
                                    $md5 = md5_base64($val);
                                    $out .= "<value>$md5</value>";
                                }
                            }
                        }
                        else {
                            foreach $val (@values) {
                                $md5 = md5_base64($val);

                                #AOSICFG::write_ws_log("| ldapGenSearch(multi): md5:$md5;");
                                if ($md5_in eq $md5) {
                                    $out .= "<value>" . encode_base64($val) . "</value>";
                                }
                            }
                        }
                    }
                    else {
                        foreach $val (@values) {
                            AOSICFG::write_ws_log("| $log: val=$val;") if $config::aosi_debug_level > 3;
                            $temp = "<value>" . XML::Simple::escape_value($dummy, $val) . "</value>";
                            Encode::_utf8_on($temp);
                            $out .= $temp;
                            AOSICFG::write_ws_log("| $log: out=$out;") if $config::aosi_debug_level > 3;

                        }
                    }
                    $out .= "</attribute>";
                }
                AOSICFG::write_ws_log("| $log: XML building finished") if $config::aosi_debug_level > 3;
                ($plugin_hash) = Plugins::bonusSearch($service, $user, $dn, \@attrs);
                ($plugin_msg, $plugin_code, $plugin_xml) = AOSICFG::process_plugin_hash($plugin_hash, 'search');
                $out .= $plugin_xml;
                $out .= "</entry>";
            }
            if ($from > -1) {
                $hash->{'from'}  = $from;
                $hash->{'size'}  = $size;
                $hash->{'count'} = $#entries + 1;
            }
            $out              = "<?xml version=\"1.0\"?><ldap action=\"$service\" version=\"" . $config::aosi_version . "\">$out</ldap>";
            $hash->{'result'} = $out;
            $hash->{'error'}  = '';
        }
        else {
            $hash->{'error'}  = $mesg2->error;
            $hash->{'result'} = '';
        }
        $mesg2 = $ldap->unbind;    # take down session

    }
    else {
        $hash->{'code'}  = $auth_code;
        $hash->{'error'} = $auth_error;
		$pp_status = '';
		if($pp_mesg){
			$pp_status = ' pwdPolicySubEntry="'.$pp_mesg.'"';
		}
        $hash->{'result'} =
            '<?xml version="1.0"?><ldap action="' . $service
          . '" version="'
          . $config::aosi_version
          . '"><entry dn="'
          . $dn
          . '"><status code="'
        . $auth_code . '"'.$pp_status.'>'
          . $auth_error
          . '</status></entry></ldap>';
    }

    #$hash_val ='';
    AOSICFG::write_search_log($user, $service, $base, $filter, $attributes, $code);
    AOSICFG::write_ws_log("| $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 2;
    return $hash;
}

sub bydn {
    $a->dn() cmp $b->dn();
}

######################################################################
# Generic subroutine for attribute manipulation
######################################################################
sub ldapGenModify {
    my ($service, $user, $pwd, $base, $xml_in, $type, $bind_user) = @_;
    my ($att,         $attr_exists, $attrs,    $auth_code, $auth_error, $bin_attr_ex);
    my ($bin_ent,     $bin_error,   $bin_hash, $bin_msg,   $bin_val,    $bin_xml, $code, $cur_code, $cur_md5);
    my ($cur_old_md5, $cur_val,     $dec_pwd,  $dn,        $dn_in,      $dummy, $ent, $entries, $err_msg, $err_num);
    my ($hash,        $hostname,    $i,        $is_admin,  $j,          $lbl, $ldap, $len, $log, $md5, $mesg2);
    my ($mod_hash, $old, $old_bin, $old_val,    $out_msg, $port,    $prev_val, $realm,    $temp_msg);
    my ($uid,      $val, %md5_del, %md5_exists, @del,     @entries, @mod_keys, @new_vals, @old_md5);
    my (@old_vals, @to_del, @vals, @hep_af_vals, %options, $tmp_val, $all_attributes, $hepun, $tmp_code, $tmp_msg, $hepun2);
    my ($admin_filter, $my_filter, @entries2, $mesg3, $ldap_error, $plugin_code, $plugin_msg, $out_one, $OIB);
	my ($restriction_code, $restriction_msg, $ii, $iji, $ik);
    $log = 'ldapGenModify';
    AOSICFG::write_ws_log("| $log(in):user:$user;base:$base;xml:$xml_in;type:$type;bind_user:$bind_user") if $config::aosi_debug_level > 2;

    ($hostname, $port)       = parse_server($config::ldap_server);
    ($ldap,     $ldap_error) = ldap_init($hostname, $port);
    ($uid,      $realm)      = split(/\@/, $user);
    $dn      = 'uid=' . $uid . ',' . $base;
    $code    = 0;
    $err_num = 0;
    $out_msg = '';

    if ($bind_user == 1) {
        ($auth_code, $auth_error) = ck_auth($ldap, get_dn($user, $base), $pwd, 0, $ldap_error);
    }
    else {
        ($auth_code, $auth_error, $is_admin, $admin_filter) = ck_auth($ldap, get_dn($user, $base), $pwd, 1, $ldap_error);
        Encode::_utf8_on($admin_filter);
        if ($is_admin == 0 and $ldap) {
            if ($auth_code == 0) {
                $auth_code  = -15;
                $auth_error = "User $user does not have admin privileges!";
            }
        }

    }
    if ($auth_code == 0) {
        if ($bind_user == 1) {
            $dec_pwd = decode_base64($pwd);
            $dn      = get_dn($user, $base);
            $mesg2   = $ldap->bind($dn, password => $dec_pwd, version => 3);
            AOSICFG::write_ws_log("| $log(bind_as_user): " . $mesg2->code) if $config::aosi_debug_level > 2;
        }
        else {
            $mesg2 = $ldap->bind($config::ldap_user, password => AOSICFG::get_pwd($config::ldap_pwd_file), version => 3);
            AOSICFG::write_ws_log("| $log(bind_as_admin): " . $mesg2->code) if $config::aosi_debug_level > 2;
            if ($mesg2->code != 0) {
                AOSICFG::write_ws_log("| !!!$log: invalid admin password in " . $config::ldap_pwd_file);
            }
        }
        ($entries, $attr_exists) = xml2entries($xml_in);
        @entries = @{$entries};

        foreach $ent (@entries) {
            $dn_in = AOSICFG::utfize($ent->{'dn'});
            ($uid, $dummy) = split(/,/, $dn_in);
            ($lbl, $uid)   = split(/=/, $uid);
            $out_one = '<entry dn="' . $dn_in . '">';

            #$out_msg .= $out_one;
            $my_filter = '(&(uid=' . $uid . ')' . $admin_filter . ')';
            AOSICFG::write_ws_log("| $log:my_filter:$my_filter") if $config::aosi_debug_level > 2;
            $mesg3    = $ldap->search(base => $base, filter => $my_filter, timelimit => 60, attrs => ['uid']);
            $cur_code = $mesg3->code;
            @entries2 = $mesg3->entries;
            AOSICFG::write_ws_log("| $log(search with my_filter):code:$cur_code;error:" . $mesg3->error . ";entries:" . $#entries2)
              if $config::aosi_debug_level > 2;
            if ($cur_code == 0 and $#entries2 > -1) {
                ($plugin_hash) = Plugins::beforeChangeAttribute($type, $ldap, $base, $ent, $user, $pwd);
                ($plugin_panic, $plugin_code, $plugin_xml) = AOSICFG::process_plugin_hash($plugin_hash, 'before');
                $out_one .= $plugin_xml;
                if ($plugin_panic == 0) {
                    $attrs = $ent->{'attrs'};
                    $len   = $#{$attrs};

                    #@del = ();
                    #@vals = ();
                    $mod_hash = undef;
                    foreach $i (0 .. $len) {
                        @del  = ();
                        @vals = ();
                        $att  = $attrs->[$i]->{'att'};
                        AOSICFG::write_ws_log("| $log(===== processing $att ====)") if $config::aosi_debug_level > 3;
                        if ($att eq 'phoneNumber') {
                            $att = 'homeTelephoneNumber';
                        }
                        $val = $attrs->[$i]->{'val'};
                        if (ref($val) eq 'ARRAY') {
                            @vals = @$val;
                            if ($#vals == -1) {
                                $vals[0] = '';
                            }
                            AOSICFG::write_ws_log("| $log($att; multi:" . $#vals . "): " . join(", ", @vals))
                              if ($config::aosi_debug_level > 2 and $att ne 'userPassword');
                        }
                        else {
                            if ($val) {
                                push(@vals, $val);
                            }
                            AOSICFG::write_ws_log("| $log($att; single): $val") if ($config::aosi_debug_level > 2 and $att ne 'userPassword');
                        }
                        if (AOSICFG::is_bin($att)) {
                            if ($bind_user == 1) {
                                $bin_hash = ldapGenSearch($service, $user, $pwd, $base, "(uid=$uid)", $att, 1, -1, 0, 1, $md5);
                            }
                            else {
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", $att, 0, -1, 0, 1, $md5);
                            }

                            $bin_xml = $bin_hash->{'result'};
                            ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                            $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                            if (ref($bin_val) eq 'ARRAY') {    # Multi
                                @old_vals = @$bin_val;
                            }
                            else {                             # Single
                                if ($bin_val) {
                                    @old_vals = ($bin_val);
                                }
                            }
                            AOSICFG::write_ws_log("| $log(old_multi:" . $#old_vals . "): " . join(", ", @old_vals)) if $config::aosi_debug_level > 2;
                            if ($type == 1) {                  # Add
                                AOSICFG::write_ws_log("| $log(bin_add): ")               if $config::aosi_debug_level > 2;
                                AOSICFG::write_ws_log("| $log(old_vals): " . $#old_vals) if $config::aosi_debug_level > 2;
                                if ($#old_vals > -1) {
                                    @old_md5 = ();
                                    foreach $old_val (@old_vals) {
                                        $old_bin     = decode_base64($old_val);
                                        $cur_old_md5 = md5_base64($old_val);
                                        $md5_exists{$cur_old_md5}++;
                                        push(@{ $mod_hash->{$att} }, $old_bin);
                                    }
                                }
                                AOSICFG::write_ws_log("| $log(vals): " . $#vals) if $config::aosi_debug_level > 2;
                                foreach $cur_val (@vals) {
                                    $cur_md5 = md5_base64($cur_val);
                                    if (!$md5_exists{$cur_md5}) {
                                        push(@{ $mod_hash->{$att} }, decode_base64($cur_val));
                                    }
                                }
                                $bin_error = 0;
                            }
                            elsif ($type == 2) {    # Modify
                                AOSICFG::write_ws_log("| $log(mod_bin): $md5") if $config::aosi_debug_level > 2;
                                $bin_error = -31;
                                $bin_msg   = "Modification of binary attribute $att not allowed!";
                            }
                            elsif ($type == 3) {    # Delete
                                AOSICFG::write_ws_log("| $log(del_bin): $md5") if $config::aosi_debug_level > 2;
                                AOSICFG::write_ws_log("| $log(bin_vals): (" . join(",", @vals) . ")") if $config::aosi_debug_level > 2;
                                foreach $val (@vals) {
                                    $md5_del{$val}++;
                                }
                                if ($#old_vals > -1 and $#vals > -1) {
                                    @old_md5 = ();
                                    foreach $old_val (@old_vals) {
                                        $old_bin     = decode_base64($old_val);
                                        $cur_old_md5 = md5_base64($old_bin);
                                        push(@old_md5, $cur_old_md5);
                                        if (!$md5_del{$cur_old_md5}) {
                                            push(@{ $mod_hash->{$att} }, $old_bin);
                                        }
                                    }
                                    AOSICFG::write_ws_log("| $log(old_md5):  (" . join(",", @old_md5) . ")") if $config::aosi_debug_level > 2;
                                }
                                $bin_error = 0;
                                AOSICFG::write_ws_log("| $log(mod_hash2):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;

                                #AOSICFG::write_ws_log("| $log(del_bin): $val");
                                #$mesg2 = $ldap->modify( $dn_in,  delete => [$att]);
                            }
                        }
                        else {
                            AOSICFG::write_ws_log("| $log: $att not_bin") if $config::aosi_debug_level > 2;
                            AOSICFG::write_ws_log("| $log($att; not_bin): " . join(", ", @vals)) if ($config::aosi_debug_level > 2 and $att ne 'userPassword');
                            if ($#vals > -1) {
                                $restriction_code = 0;
                                foreach $j (0 .. $#vals) {
                                    $val = $vals[$j];
                                    AOSICFG::write_ws_log("| $log(cur_val): $val") if ($config::aosi_debug_level > 2 and $att ne 'userPassword');
                                    if ($val eq '' and $type == 3) {
                                        push(@del, $att);
                                        AOSICFG::write_ws_log("| $log($att; del_all_1)") if $config::aosi_debug_level > 2;
                                    }
                                    else {
                                        #if($j == 0){
                                        #    $mod_hash->{$att} = $val;
                                        #}
                                        #elsif ($j == 1){
                                        #    $prev_val = $mod_hash->{$att};
                                        #    $mod_hash->{$att} = [$prev_val, $val];
                                        #}
                                        #elsif($j > 1){
                                        #    push(@{$mod_hash->{$att}}, $val);
                                        #}
                                        ($cur_code, $temp_msg) = is_restriction_ok($att, $val);
                                        if (($cur_code == 1) or ($type == 3)){
                                            push(@{ $mod_hash->{$att} }, $val);
                                            $restriction_code += 0;
                                        }
                                        else {
                                            if($att eq 'hrEduPersonUniqueNumber'){
                                                $restriction_code += -55;
                                                $restriction_msg = $temp_msg;
                                            }
                                            else {
                                                $restriction_code += -63;
                                                $restriction_msg = $temp_msg;
                                            }
                                        }
                                        AOSICFG::write_ws_log("| $log(restriction): code=$restriction_code;msg=" . $restriction_msg)
                                          if $config::aosi_debug_level > 2;
                                        AOSICFG::write_ws_log("| $log(mod_not_bin):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    }
                                    #$out_msg .= "$att;";
                                }
                            }
                            else {
                                if ($type == 3) {
                                    push(@del, $att);
                                    AOSICFG::write_ws_log("| $log($att; del_all_2)") if $config::aosi_debug_level > 2;
                                }
                            }
                        }
                    }
                    push(@$attrs, objectClass => 'hrEduPerson');
                    #push(@$attrs, shacObjectClass => 'schacContactLocation');
                    if (AOSICFG::is_bin($att)) {
                        if (!$bin_error) {
                            if ($type == 1) {    # Add
                                if ($#old_vals > -1) {
                                    $mesg2 = $ldap->modify($dn_in, delete => [$att]);
                                    AOSICFG::write_ws_log("| $log(bin_deleting 1): $att=" . $mesg2->error . "(" . $mesg2->code . ")")
                                      if $config::aosi_debug_level > 2;
                                    $cur_code = $mesg2->code;
                                }
                                else {
                                    AOSICFG::write_ws_log("| $log(bin_deleting 1a): $att=nothing to delete (old_vals:" . $#old_vals . ")")
                                      if $config::aosi_debug_level > 2;
                                    $cur_code = 0;
                                }
                                if ($cur_code == 0) {
                                    AOSICFG::write_ws_log("| $log(bin_adding a): $dn_in")                          if $config::aosi_debug_level > 2;
                                    AOSICFG::write_ws_log("| $log(bin_adding b):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    $mesg2 = $ldap->modify($dn_in, add => $mod_hash);
                                    $cur_code = $mesg2->code;
                                    my $all_attrs = join(", ", keys %$mod_hash);
                                    AOSICFG::write_ws_log("| $log(bin_adding): $att=" . $mesg2->error . "(" . $code . ")") if $config::aosi_debug_level > 2;
                                    if ($cur_code == 0) {
                                        $tmp_msg = '<status code="0">Added binary: ' . $all_attrs . '!</status>';
                                    }
                                    else {
                                        $tmp_msg = '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                        $err_num++;
                                    }
                                }
                                else {
                                    $tmp_msg = '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                    $err_num++;
                                }
                                $out_one .= $tmp_msg;

                                #$out_msg .= $out_one;
                            }
                            elsif ($type == 3) {    # delete
                                @to_del = ();
                                foreach $old (@old_md5) {
                                    if ($md5_del{$old}) {
                                        push(@to_del, $old);
                                    }
                                }
                                if ($#to_del > -1) {
                                    $mesg2 = $ldap->modify($dn_in, delete => [$att]);
                                    AOSICFG::write_ws_log("| $log(deleting): $att=" . $mesg2->code);
                                    $out_one .= '<status code="0">Deleted binary: $att!</status>';

                                    #$out_msg .= $out_one;
                                    if ($mod_hash->{$att}) {
                                        $mesg2 = $ldap->modify($dn_in, add => $mod_hash);
                                        AOSICFG::write_ws_log("| $log(modifying): $att=" . $mesg2->code);
                                    }
                                }
                                else {
                                    $bin_error = -32;
                                    $bin_msg   = "Nothing to delete!";
                                }
                            }
                        }
                    }
                    else {
                        if ($att eq 'homePostalAddress' or $att eq 'facsimileTelephoneNumber') {
                            if ($bind_user == 1) {
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", $att, 1, -1, 0);
                            }
                            else {
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", $att, 0, -1, 0);
                            }
                            AOSICFG::write_ws_log("| $log(hPA_old_vals):" . AOSICFG::dump_hash($bin_hash)) if $config::aosi_debug_level > 2;
                            $bin_xml = $bin_hash->{'result'};
                            ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                            $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                            if ($bin_val) {
                                @old_vals = @$bin_val;
                            }
                            else {
                                @old_vals = ();
                            }
                            if ($type == 1) {    # Add
                                foreach $old_val (@old_vals) {
                                    push(@new_vals, $old_val);
                                }
                                foreach $cur_val (@{ $mod_hash->{$att} }) {
                                    push(@new_vals, $cur_val);
                                }
                                AOSICFG::write_ws_log("| $log(hPA_add): (" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                                $dummy = 'Added';
                            }
                            elsif ($type == 2) {    # Modify
                                foreach $cur_val (@{ $mod_hash->{$att} }) {
                                    push(@new_vals, $cur_val);
                                }
                                AOSICFG::write_ws_log("| $log(hPA_mod): (" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                                $dummy = 'Modified';
                            }
                            elsif ($type == 3) {    # delete
                                foreach $old_val (@old_vals) {
                                    if ($old_val ne $val) {
                                        push(@new_vals, $old_val);
                                    }
                                }
                                if ($#old_vals > -1 and ($#old_vals != $#new_vals)) {
                                    $bin_error = 0;
                                    AOSICFG::write_ws_log("| $log(hPA_del): (" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                                    $dummy = 'Deleted';
                                }
                                else {
                                    $bin_error = -71;
                                    AOSICFG::write_ws_log("| $log(hPA_del): nothing to delete.") if $config::aosi_debug_level > 2;
                                    $bin_msg = 'Nothing to delete!';
                                }
                            }
                            if ($bin_error == 0) {
                                AOSICFG::write_ws_log("| $log(hPA_new_vals):(" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                                $mod_hash->{$att} = \@new_vals;
                                AOSICFG::write_ws_log("| $log(mod_hPA):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                $mesg2 = $ldap->modify($dn_in, delete => [$att]);
                                AOSICFG::write_ws_log("| $log(mod_hPA del):code:" . $mesg2->code . ";error:" . $mesg2->error) if $config::aosi_debug_level > 3;
                                $bin_error = $mesg2->code;
                                $bin_msg   = $mesg2->error;
                                $mesg2     = $ldap->modify($dn_in, replace => $mod_hash);
                                $bin_error = $mesg2->code;
                                $bin_msg   = $mesg2->error;
                                AOSICFG::write_ws_log("| $log(mod_hPA replace):code:" . $mesg2->code . ";error:" . $mesg2->error)
                                  if $config::aosi_debug_level > 3;
                            }
                            if ($bin_error == 0) {
                                $out_one .= '<status code="0">' . $dummy . ' attrbiute ' . $att . '!</status>';

                                #$out_msg .= $out_one;
                            }
                            else {
                                $mod_hash->{$att} = \@old_vals;
                                $mesg2 = $ldap->modify($dn_in, replace => $mod_hash);
                                if ($mesg2->code != 0) {
                                    $bin_error = $mesg2->code;
                                    $bin_msg   = $mesg2->error;
                                    $code += abs($bin_error);
                                    $err_num++;
                                    $out_one .= '<status code="' . $bin_error . '">' . $bin_msg . '</status>';

                                    #$out_msg .= $out_one;
                                }
                                AOSICFG::write_ws_log("| $log(mod_hPA put old vals back):code:" . $mesg2->code . ";error:" . $mesg2->error)
                                  if $config::aosi_debug_level > 3;
                            }
                        }
                        elsif ($att eq 'hrEduPersonPrimaryAffiliation') {
                            if ($type == 1) {    # Add
                                $out_one .= '<status code="-41">Attribute ' . $att . ' cannot be added! Add a value to hrEduPersonAffiliation!</status>';

                                #$out_msg .= $out_one;
                                $code = 41;
                                $err_num++;
                            }
                            elsif ($type == 3) {    # Delete
                                $out_one .=
                                    '<status code="-42">Attribute ' . $att
                                  . ' cannot be deleted! In order to delete it you have to delete hrEduPersonAffiliation!</status>';

                                #$out_msg .= $out_one;
                                $code = 43;
                                $err_num++;
                            }
                            elsif ($type == 2) {    # Modify
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", 'hrEduPersonAffiliation', 0, -1, 0);
                                $bin_xml  = $bin_hash->{'result'};
                                $cur_code = $bin_hash->{'code'};
                                ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                if (ref($bin_val) eq 'ARRAY') {
                                    @hep_af_vals = @$bin_val;
                                }
                                else {
                                    @hep_af_vals = ($bin_val);
                                }
                                AOSICFG::write_ws_log("| $log(hep_PAf get hep_Af): code:$cur_code;values:[" . join(",", @hep_af_vals) . "]")
                                  if $config::aosi_debug_level > 2;
                                if ($cur_code == 0) {
                                    if ($#hep_af_vals > -1) {
                                        $dummy   = 0;
                                        $cur_val = @{ $mod_hash->{$att} }[0];
                                        foreach $tmp_val (@hep_af_vals) {
                                            if ($tmp_val eq $cur_val) {
                                                $dummy++;
                                            }
                                        }
                                        AOSICFG::write_ws_log("| $log(hep_PAf find equal in hep_Af): $dummy ($cur_val)") if $config::aosi_debug_level > 2;
                                        if ($dummy > 0) {
                                            $mesg2 = $ldap->modify($dn_in, replace => $mod_hash);
                                            $cur_code = $mesg2->code;
                                            if ($cur_code == 0) {
                                                $code = $cur_code;
                                                $out_one .= '<status code="0">Modified attribute ' . $att . '!</status>';

                                                #$out_msg .= $out_one;
                                            }
                                            else {
                                                $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';

                                                #$out_msg .= $out_one;
                                                $code += $cur_code;
                                                $err_num++;
                                            }
                                        }
                                        else {
                                            $cur_code = -43;
                                            $out_one .=
                                                '<status code="'
                                              . $cur_code
                                              . '">hrEduPersonPrimaryAffiliation can be set only to one of the values in hrEduPersonAffiliation!</status>';

                                            #$out_msg .= $out_one;
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                    }
                                    else {
                                        $cur_code = -44;
                                        $out_one .=
                                            '<status code="'
                                          . $cur_code
                                          . '">Attribute hrEduPersonAffiliation is empty! Add a value to this attribute first!</status>';

                                        #$out_msg .= $out_one;
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                                else {
                                    $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';

                                    #$out_msg .= $out_one;
                                    $code += $cur_code;
                                    $err_num++;
                                }
                            }
                        }
                        elsif ($att eq 'hrEduPersonAffiliation') {
                            if ($restriction_code == 0) {
                                if ($type == 1) {    # Add
                                    AOSICFG::write_ws_log("| $log(add):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    $mesg2 = $ldap->modify($dn_in, add => $mod_hash);
                                    $cur_code = $mesg2->code;
                                    if ($cur_code == 0) {
                                        my $all_attrs = join(", ", keys %$mod_hash);
                                        $out_one .= '<status code="0">Added attribute ' . $all_attrs . '!</status>';

                                        #$out_msg .= $out_one;
                                    }
                                    else {
                                        $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';

                                        #$out_msg .= $out_one;
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                                elsif ($type == 3 or $type == 2) {    # Delete or Modify
                                    $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", 'hrEduPersonPrimaryAffiliation', 0, -1, 0);
                                    $bin_xml  = $bin_hash->{'result'};
                                    $cur_code = $bin_hash->{'code'};
                                    ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                    $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                    if (ref($bin_val) eq 'ARRAY') {
                                        $bin_val = $bin_val->[0];
                                    }
                                    AOSICFG::write_ws_log("| $log(hep_Af get hep_PAf): code:$cur_code;value:$bin_val") if $config::aosi_debug_level > 2;
                                    $dummy       = 0;
                                    @hep_af_vals = @{ $mod_hash->{$att} };
                                    foreach $tmp_val (@hep_af_vals) {
                                        if ($tmp_val eq $bin_val) {
                                            $dummy++;
                                        }
                                    }
                                    AOSICFG::write_ws_log("| $log(hep_Af find equal in hep_PAf): $dummy ($bin_val) in [" . join(",", @hep_af_vals) . "]")
                                      if $config::aosi_debug_level > 2;
                                    if ($type == 3) {    # Delete
                                        if ($dummy == 0) {
                                            AOSICFG::write_ws_log("| $log(hep_Af delete):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                            $mesg2 = $ldap->modify($dn_in, delete => $mod_hash);
                                            $cur_code = $mesg2->code;
                                            if ($cur_code == 0) {
                                                $code = $cur_code;
                                                $out_one .= '<status code="0">Deleted attribute ' . $att . '!</status>';

                                                #$out_msg .= $out_one;
                                            }
                                            else {
                                                $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';

                                                #$out_msg .= $out_one;
                                                $code += $cur_code;
                                                $err_num++;
                                            }
                                        }
                                        else {
                                            $cur_code = -51;
                                            $out_one .=
                                                '<status code="'
                                              . $cur_code
                                              . '">This value cannot be deleted, hrEduPersonPrimaryAffiliation is set to that value!</status>';

                                            #$out_msg .= $out_one;
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                    }
                                    elsif ($type == 2) {    # Modify
                                        if ($dummy == 1) {
                                            AOSICFG::write_ws_log("| $log(hep_Af delete):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                            $mesg2 = $ldap->modify($dn_in, replace => $mod_hash);
                                            $cur_code = $mesg2->code;
                                            if ($cur_code == 0) {
                                                $code = $cur_code;
                                                $out_one .= '<status code="0">Modified attribute ' . $att . '!</status>';

                                                #$out_msg .= $out_one;
                                            }
                                            else {
                                                $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';

                                                #$out_msg .= $out_one;
                                                $code += $cur_code;
                                                $err_num++;
                                            }
                                        }
                                        else {
                                            $cur_code = -52;
                                            $out_one .=
                                                '<status code="'
                                              . $cur_code
                                              . '">This attribute cannot be modified to values which do not contain the value of hrEduPersonPrimaryAffiliation!</status>';

                                            #$out_msg .= $out_one;
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                    }
                                }
                            }
                            else {
                                $code += abs($restriction_code);
                                $out_one .= '<status code="' . $restriction_code . '">' . $restriction_msg . '!</status>';
                                $err_num++;
                            }
                        }
                        elsif ($att eq 'hrEduPersonOIB') {    # Special hamndling of hrEduPersonOIB - it tied to hrEduPersonUniqueNumber
                            if ($restriction_code == 0) {
                                AOSICFG::write_ws_log("| $log(hrEduPersonOIB): val=".join("", @vals).";type=$type") if $config::aosi_debug_level > 2;
                                if ($type == 2) {    # Modify only - it is required and freq 1
                                    # Get the existing values in hrEduPersonUniqueNumber
                                    $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", 'hrEduPersonUniqueNumber', 0, -1, 0);
                                    $bin_xml  = $bin_hash->{'result'};
                                    $cur_code = $bin_hash->{'code'};
                                    ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                    $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                    if (ref($bin_val) eq 'ARRAY') {
                                        @hep_af_vals = @$bin_val;
                                    }
                                    else {
                                        @hep_af_vals = ($bin_val);
                                    }
                                    $tmp_code = 0;   # For handling the special cases
                                    AOSICFG::write_ws_log("| $log(hep_UN get OIB): code:$cur_code;values:[" . join(",", @hep_af_vals) . "]")
                                      if $config::aosi_debug_level > 2;
                                    $bin_val = join("", @vals);
                                    if($#hep_af_vals > -1){   # There are existing values in hrEduPersonUniqueNumber
                                        foreach $ij (0..$#hep_af_vals){
                                            $key = $hep_af_vals[$ij];
                                            if($key =~ /^OIB:\s*(\w+)/){    # There is an OIB valeu in hrEduPersonUniqueNumber
                                                if($#hep_af_vals == 0){     # The OIB value is the only existing value in hrEduPersonUniqueNumber
                                                    if($bin_val eq 'NONE'){ # No value is not NONE
                                                        $tmp_code = 1;      # hrEduPersonUniqueNumber can not be changed (e.g. deleted for NONE) since it has to have a value
                                                        AOSICFG::write_ws_log("| $log(delete OIB in hrEduPersonUniqueNumber): oib=$key;OIB is single( \#array=".$#hep_af_vals.")") if $config::aosi_debug_level > 2;
                                                    }
                                                    else {
                                                        $tmp_code = 2;      # For later action: do not delete hrEduPersonUniqueNumber but change it
                                                    }
                                                }
                                                else {    # Delete all OIB values in hrEduPersonUniqueNumber
                                                    $mesg2 = $ldap->modify($dn_in, delete => { 'hrEduPersonUniqueNumber' => $key});
                                                    $cur_code = $mesg2->code;
                                                    AOSICFG::write_ws_log("| $log(delete OIB in hrEduPersonUniqueNumber): oib=$key;code=$cur_code;error=".$mesg2->error) if $config::aosi_debug_level > 2;
                                                }
                                                splice(@hep_af_vals, $ij, 1);    # collect existing values, i.e. delete OIBs, so they can be forwarded to plug-ins
                                            }
                                        }
                                        if($bin_val ne 'NONE'){
                                            AOSICFG::write_ws_log("| $log(OIB is not NONE): bin_val=$bin_val;key=$key;") if $config::aosi_debug_level > 3;
                                            if($tmp_code == 2){   # OIB is the only value in hrEduPersonUniqueNumber so it has to be changed
                                                AOSICFG::write_ws_log("| $log(OIB is the only value)") if $config::aosi_debug_level > 3;
                                                $mesg2 = $ldap->modify($dn_in, replace => {'hrEduPersonUniqueNumber' => 'OIB: '.$bin_val});
                                                $tmp_code = 0;    # Resume normal operation
                                            }
                                            else {                # Add a new OIB value to hrEduPersonUniqueNumber
                                                $mesg2 = $ldap->modify($dn_in, add => {'hrEduPersonUniqueNumber' => 'OIB: '.$bin_val});
                                                AOSICFG::write_ws_log("| $log(Add a new OIB value to hrEduPersonUniqueNumber): dn=$dn_in;val=$bin_val;code=".$mesg2->code.";") if $config::aosi_debug_level > 3;
                                            }
                                            $cur_code = $mesg2->code;
                                            $bin_error = '';
                                            if($cur_code != 0){
                                                $bin_error = $mesg2->error;
                                            }
                                            foreach $ii(0..$#{$ent->{'attrs'}}){
                                                $dummy = $ent->{'attrs'}->[$ii];
                                                if($dummy->{'att'} eq 'hrEduPersonUniqueNumber'){
                                                      splice(@{$ent->{'attrs'}}, $ii);
                                                      AOSICFG::write_ws_log("| $log(deleting hrEduPersonUniqueNumber): ii=$ii;dummy=$dummy;att_name=".$dummy->{'att'}.";") if $config::aosi_debug_level > 3;
                                                }
                                            }
                                            push(@hep_af_vals, 'OIB: '.$bin_val);   # Final array off all values for hrEduPersonUniqueNumber forwarded to plug-ins
                                            push(@{$ent->{'attrs'}}, {'att' => 'hrEduPersonUniqueNumber', 'val' => \@hep_af_vals});
                                        }
                                        else {  # OIB is set to NONE so other hrEduPersonUniqueNumber values are forwarded to plug-ins
                                            push(@{$ent->{'attrs'}}, {'att' => 'hrEduPersonUniqueNumber', 'val' => \@hep_af_vals});
                                        }
                                    }
                                    if($mod_hash->{'hrEduPersonUniqueNumber'}){
                                        if($bin_val ne 'NONE'){
                                            push(@{$mod_hash->{'hrEduPersonUniqueNumber'}}, 'OIB: '.$bin_val);
                                            AOSICFG::write_ws_log("| $log(hrEduPersonUniqueNumber is handled in same request): OIB=$bin_val;mod_hash:".dump_hash($mod_hash)) if $config::aosi_debug_level > 3;
                                        }
                                    }
                                    if($tmp_code == 0){
                                        $mesg2 = $ldap->modify($dn_in, replace => $mod_hash);
                                        $cur_code = $mesg2->code;
                                        if ($cur_code == 0) {
                                            $temp_msg = $bin_error eq '' ? 'hrEduPersonUniqueNumber OIB value is also updated' : 'hrEduPersonUniqueNumber OIB value could not be updated: '.$bin_error;
                                            $out_one .= '<status code="0">Modified attribute ' . $att . ' ('.$temp_msg.')!</status>';
                                        }
                                        else {
                                            $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                        AOSICFG::write_ws_log("| $log(modify OIB):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    }
                                    else {
                                        $cur_code = -59;
                                        $out_one .= '<status code="' . $cur_code . '">hrEduPersonOIB cannot be set to NONE because OIB is the only value in hrEduPersonUniqueNumber!</status>';
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                            }
                            else {
                                $code += abs($restriction_code);
                                $out_one .= '<status code="' . $restriction_code . '">' . $restriction_msg . '!</status>';
                                $err_num++;
                            }
                        }
                        elsif($att eq 'hrEduPersonUniqueNumber'){
                            if($restriction_code == 0){
                                $cur_code = 0;
                                $bin_msg = '';
                                $tmp_code = 0;   # For handling the special cases
                                foreach $hepun (@{$mod_hash->{'hrEduPersonUniqueNumber'}}){
                                    AOSICFG::write_ws_log("| $log: hrEduPersonUniqueNumber: val=$hepun;") if $config::aosi_debug_level > 3;
                                    if($hepun =~ /^OIB:.*/){  # OIB stuff
                                        if(AOSICFG::aosi_version(2) > 2){
                                            $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", 'hrEduPersonOIB', 0, -1, 0);
                                            $bin_xml  = $bin_hash->{'result'};
                                            $cur_code = $bin_hash->{'code'};
                                            ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                            $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                            if (ref($bin_val) eq 'ARRAY') {
                                                @hep_af_vals = @$bin_val;
                                            }
                                            else {
                                                @hep_af_vals = ($bin_val);
                                            }
                                            $OIB = join("", @hep_af_vals);
                                            AOSICFG::write_ws_log("| $log:hrEduPersonOIB=$OIB") if $config::aosi_debug_level > 3;
                                        }
                                        if ($type == 1) {       # Add
                                            # Check if OIB allready exists
                                            $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", 'hrEduPersonUniqueNumber', 0, -1, 0);
                                            $bin_xml  = $bin_hash->{'result'};
                                            $cur_code = $bin_hash->{'code'};
                                            ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                            $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                            if (ref($bin_val) eq 'ARRAY') {
                                                @hep_af_vals = @$bin_val;
                                            }
                                            else {
                                                @hep_af_vals = ($bin_val);
                                            }
                                            foreach $bin_val(@hep_af_vals){
                                                if($bin_val =~ /^OIB:.*/){
                                                    $cur_code = -57;
                                                    $bin_msg = 'OIB value for hrEduPersonUniqueNumber allready exists!';
                                                }
                                            }
                                        }
                                        elsif ($type == 2) {    # Modify
                                            if(AOSICFG::aosi_version(2) > 2){
                                                foreach $hepun2 (@{$mod_hash->{'hrEduPersonUniqueNumber'}}){
                                                    if($hepun2 =~ /^OIB:\s*(\d+)/){
                                                        if($hepun2 ne 'OIB: '.$OIB){
                                                            AOSICFG::write_ws_log("| $log:hrEduPersonUniqueNumber($hepun) ? hrEduPersonOIB($OIB)") if $config::aosi_debug_level > 3;
                                                            $cur_code = -56;
                                                            $bin_msg = 'OIB value in hrEduPersonUniqueNumber has to be equal to the value of hrEduPersonOIB!';
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        elsif ($type == 3) {    # Delete
                                            $cur_code = -58;
                                            $bin_msg = 'OIB value of hrEduPersonUniqueNumber cannot be deleted!';
                                        }
                                    }
                                    else {  # This is not OIB that is added
                                        # Check if OIB allready exists and if it is NONOE
                                        $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", 'hrEduPersonUniqueNumber', 0, -1, 0);
                                        $bin_xml  = $bin_hash->{'result'};
                                        $cur_code = $bin_hash->{'code'};
                                        ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                        $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                        if (ref($bin_val) eq 'ARRAY') {
                                            @hep_af_vals = @$bin_val;
                                        }
                                        else {
                                            @hep_af_vals = ($bin_val);
                                        }
                                        foreach $bin_val(@hep_af_vals){
                                            if($bin_val =~ /^OIB:.*NONE/){  # Check if there is OIB: NONE in hrEduPersonUniqueNumber
                                                $tmp_code = 1;
                                            }
                                        }
                                    }
                                }
                                AOSICFG::write_ws_log("| $log: check hrEduPersonUniqueNumber out: code=$cur_code;msg=$bin_msg;") if $config::aosi_debug_level > 3;
                                if($cur_code == 0){
                                    if($mod_hash->{'hrEduPersonOIB'}){
                                        $OIB = join("", @{$mod_hash->{'hrEduPersonOIB'}});
                                        push(@{$mod_hash->{'hrEduPersonUniqueNumber'}}, 'OIB: '.$OIB);
                                        AOSICFG::write_ws_log("| $log(OIB is handled in same request): OIB=$OIB;mod_hash:".dump_hash($mod_hash)) if $config::aosi_debug_level > 3;
                                    }
                                    if ($type == 1) {         # Add
                                        $mesg2 = $ldap->modify($dn_in, add => $mod_hash);
                                        $cur_code = $mesg2->code;
                                        if ($cur_code == 0) {
                                            if($tmp_code == 1){   # If there is OIB: NONE in hrEduPersonUniqueNumber we can delete it now (since it is not the only value)
                                                $mesg2 = $ldap->modify($dn_in, delete => { 'hrEduPersonUniqueNumber' => 'OIB: NONE'});
                                                $cur_code = $mesg2->code;
                                                AOSICFG::write_ws_log("| $log(delete OIB: NONE in hrEduPersonUniqueNumber): code=$cur_code;error=".$mesg2->error) if $config::aosi_debug_level > 2;
                                            }
                                            my $all_attrs = join(", ", keys %$mod_hash);
                                            $out_one .= '<status code="0">Added attribute hrEduPersonUniqueNumber!</status>';
                                        }
                                        else {
                                            $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                    }
                                    elsif ($type == 2) {    # Modify
                                        $mesg2 = $ldap->modify($dn_in, replace => $mod_hash);
                                        $cur_code = $mesg2->code;
                                        if ($cur_code == 0) {
                                            $out_one .= '<status code="0">Modified attribute hrEduPersonUniqueNumber!</status>';
                                        }
                                        else {
                                            $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                    }
                                    elsif ($type == 3) {    # Delete
                                        $mesg2 = $ldap->modify($dn_in, delete => $mod_hash);
                                        $cur_code = $mesg2->code;
                                        if ($cur_code == 0) {
                                           $out_one .= '<status code="0">Deleted attribute hrEduPersonUniqueNumber!</status>';
                                        }
                                        else {
                                           $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                           $code += abs($cur_code);
                                           $err_num++;
                                        }
                                    }
                                }
                                else {
                                    $out_one .= '<status code="' . $cur_code . '">'.$bin_msg.'</status>';
                                    $code += abs($cur_code);
                                    $err_num++;
                                }
                            }
                            else {
                                $code += abs($restriction_code);
                                $out_one .= '<status code="' . $restriction_code . '">' . $restriction_msg . '!</status>';
                                $err_num++;
                            }
                        }
                        elsif ($att eq 'ou' and $admin_filter ne '') {
                            if ($restriction_code == 0) {
                                my @raw = split(/\)\(/, $admin_filter);
                                my $ou_ok = 0;
                                my (@final);
                                foreach my $ou (@raw) {
                                    $ou =~ s/[\)\(\|\&]//g;
                                    next if $ou !~ /^ou=/;
                                    $ou =~ s/ou=//g;
                                    push(@final, $ou);
                                    if ($ou eq $val) {
                                        $ou_ok++;
                                    }
                                }
                                AOSICFG::write_ws_log("| $log(ou): att=$att;val=$val;admin_filter=$admin_filter;ou_ok=$ou_ok;final=" . join(", ", @final))
                                  if $config::aosi_debug_level > 2;
                                if ($ou_ok > 0) {
                                    if ($type == 1) {    # Add
                                        AOSICFG::write_ws_log("| $log(add ou):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                        $mesg2 = $ldap->modify($dn_in, add => $mod_hash);
                                        $cur_code = $mesg2->code;
                                        if ($cur_code == 0) {
                                            my $all_attrs = join(", ", keys %$mod_hash);
                                            $out_one .= '<status code="0">Added attribute ' . $all_attrs . '!</status>';
                                        }
                                        else {
                                            $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                    }
                                    elsif ($type == 2) {    # Modify
                                        AOSICFG::write_ws_log("| $log(mod ou):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                        $mesg2 = $ldap->modify($dn_in, replace => $mod_hash);
                                        $cur_code = $mesg2->code;
                                        if ($cur_code == 0) {
                                            my $all_attrs = join(", ", keys %$mod_hash);
                                            $out_one .= '<status code="0">Modified attribute ' . $all_attrs . '!</status>';
                                        }
                                        else {
                                            $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                    }
                                    elsif ($type == 3) {    # delete
                                        $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", $att, 0, -1, 0);
                                        $bin_xml  = $bin_hash->{'result'};
                                        $cur_code = $bin_hash->{'code'};
                                        ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                        $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                        my ($bin_num);
                                        if (ref($bin_val) eq 'ARRAY') {
                                            $bin_num = $#{$bin_val};
                                        }
                                        else {
                                            $bin_num = 0;
                                        }
                                        AOSICFG::write_ws_log("| $log(old_vals ou): bin_num=$bin_num;" . AOSICFG::dump_hash($bin_hash))
                                          if $config::aosi_debug_level > 2;
                                        if ($bin_num > 0) {    # at least one ou will remain
                                            AOSICFG::write_ws_log("| $log(del ou):" . join(",", @del)) if $config::aosi_debug_level > 2;
                                            my ($all_attrs);
                                            if ($#del > -1) {
                                                $all_attrs = join(", ", @del);
                                                $mesg2 = $ldap->modify($dn_in, delete => \@del);
                                            }
                                            @mod_keys = keys %$mod_hash;
                                            if ($#mod_keys > -1) {
                                                $tmp_val = $all_attrs eq '' ? '' : ',';
                                                $all_attrs .= $tmp_vals . join(", ", keys %$mod_hash);
                                                $mesg2 = $ldap->modify($dn_in, delete => $mod_hash);
                                            }
                                            AOSICFG::write_ws_log("| $log(del_dumphash ou):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                            $cur_code = $mesg2->code;
                                            if ($cur_code == 0) {
                                                $out_one .= '<status code="0">Deleted attribute ' . $all_attrs . '!</status>';
                                            }
                                            else {
                                                $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';
                                                $code += abs($cur_code);
                                                $err_num++;
                                            }
                                        }
                                        else {
                                            $code += abs(62);
                                            $out_one .= '<status code="-62">You can not delete the last ou attribute!</status>';
                                            $err_num++;
                                        }
                                    }
                                }
                                else {
                                    $code += abs(61);
                                    $out_one .= '<status code="-61">Attribute ou has to be one from: ' . join(", ", @final) . '</status>';
                                    $err_num++;
                                }
                            }
                            else {
                                $code += abs($restriction_code);
                                $out_one .= '<status code="' . $restriction_code . '">' . $restriction_msg . '!</status>';
                                $err_num++;
                            }
                        }
                        else {
                            if ($restriction_code == 0) {
                                if ($att eq 'userPassword') {
									# Find index for userPassword
									foreach $ik(0..$len){
										if($ent->{'attrs'}->[$ik]->{'att'} eq 'userPassword'){
											$i = $ik;
										}
									}
                                    $ent->{'attrs'}->[$i]->{'val'} = encode_pwd($val);
                                    $mod_hash->{$att} = [ encode_pwd($val) ];
									# PPolicy
									if ($dn_in eq $dn) {
										$mesg2 = $ldap->modify($dn_in, delete => ['pwdPolicySubentry']);
										AOSICFG::write_ws_log("| $log(ppolicy): user changed his password - deleteing pwdPolicySubentry") if $config::aosi_debug_level > 2;                                            
									}
									else {
										$mesg2 = $ldap->modify($dn_in, add => { 'pwdPolicySubentry' => ['cn=newUser,ou=policies,'.$base]});
										AOSICFG::write_ws_log("| $log(ppolicy): admin changed users password - adding pwdPolicySubentry") if $config::aosi_debug_level > 2;
									}
                                }
                                if ($type == 1) {    # Add
                                    AOSICFG::write_ws_log("| $log(add):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    $mesg2 = $ldap->modify($dn_in, add => $mod_hash);
                                    $cur_code = $mesg2->code;
                                    if ($cur_code == 0) {
                                        my $all_attrs = join(", ", keys %$mod_hash);
                                        $out_one .= '<status code="0">Added attribute ' . $all_attrs . '!</status>';

                                        #$out_msg .= $out_one;
                                    }
                                    else {
                                        $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';

                                        #$out_msg .= $out_one;
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                                elsif ($type == 2) {    # Modify
                                    AOSICFG::write_ws_log("| $log(mod):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    $mesg2 = $ldap->modify($dn_in, replace => $mod_hash);
                                    $cur_code = $mesg2->code;
                                    if ($cur_code == 0) {
                                        my $all_attrs = join(", ", keys %$mod_hash);
                                        $out_one .= '<status code="0">Modified attribute ' . $all_attrs . '!</status>';

                                        #$out_msg .= $out_one;
                                    }
                                    else {
                                        $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';

                                        #$out_msg .= $out_one;
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                                elsif ($type == 3) {    # delete
                                    AOSICFG::write_ws_log("| $log(del):" . join(",", @del)) if $config::aosi_debug_level > 2;
                                    my ($all_attrs);
                                    if ($#del > -1) {
                                        $all_attrs = join(", ", @del);
                                        $mesg2 = $ldap->modify($dn_in, delete => \@del);
                                    }
                                    @mod_keys = keys %$mod_hash;
                                    if ($#mod_keys > -1) {
                                        $tmp_val = $all_attrs eq '' ? '' : ',';
                                        $all_attrs .= $tmp_vals . join(", ", keys %$mod_hash);
                                        $mesg2 = $ldap->modify($dn_in, delete => $mod_hash);
                                    }
                                    AOSICFG::write_ws_log("| $log(del_dumphash):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    $cur_code = $mesg2->code;
                                    if ($cur_code == 0) {
                                        $out_one .= '<status code="0">Deleted attribute ' . $all_attrs . '!</status>';

                                        #$out_msg .= $out_one;
                                    }
                                    else {
                                        $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '</status>';

                                        #$out_msg .= $out_one;
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                            }
                            else {
                                $code += abs($restriction_code);
                                $out_one .= '<status code="' . $restriction_code . '">' . $restriction_msg . '!</status>';
                                $err_num++;
                            }
                        }
                    }
                    if ($mesg2->code == 0) {
                        AOSICFG::write_ws_log("| $log(final bin_error):" . $bin_error) if $config::aosi_debug_level > 2;
                        if ($bin_error == 0) {
                            $cur_code = $mesg2->code;
                            $code += abs($cur_code);

                            #$hash->{'result'} = $out_msg;
                        }
                        else {
                            $cur_code = $bin_error;
                            $err_msg  = $bin_msg;
                            $out_one .= '<status code="' . $cur_code . '">' . $err_msg . '</status>';

                            #$out_msg .= $out_one;
                            $code += abs($cur_code);
                            $err_num++;
                        }
                    }
                    ($plugin_hash) = Plugins::afterChangeAttribute($type, $ent, $out_one . '</entry>', $user, $pwd);
                    ($plugin_panic, $plugin_code, $plugin_xml) = AOSICFG::process_plugin_hash($plugin_hash, 'after');
                    $out_one .= $plugin_xml;
                    $out_msg .= $out_one;
                    $all_attributes = join(", ", keys %$mod_hash);
                }
                else {
                    $code += abs($plugin_code);
                    $out_one .= '<status code="' . $code . '">Plugin stopped the action!</status>';
                    $out_msg .= $out_one;
                    $err_num++;
                }
            }
            else {
                $code = 83;
                $out_one .= '<status code="-' . $code . '">Attributes for user cannot be changed!</status>';
                $out_msg .= $out_one;
                $err_num++;
            }
            $out_msg .= '</entry>';
            $out_one .= '</entry>';
            AOSICFG::write_action_log($user, $service, $dn_in, $code, $all_attributes);
        }
        $hash->{'code'}   = $code;
        $hash->{'error'}  = $err_num . ' error(s) occured!';
        $hash->{'result'} = '<?xml version="1.0"?><ldap action="' . $service . '" version="' . $config::aosi_version . '">' . $out_msg . '</ldap>';
        $mesg2 = $ldap->unbind;    # take down session

    }
    else {
        $hash->{'code'}  = $auth_code;
        $hash->{'error'} = $auth_error;
        $hash->{'result'} =
            '<?xml version="1.0"?><ldap action="' . $service
          . '" version="'
          . $config::aosi_version
          . '"><entry dn="'
          . $dn
          . '"><status code="'
          . $auth_code . '">'
          . $auth_error
          . '</status></entry></ldap>';
    }
    AOSICFG::write_ws_log("| $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 2;
    return $hash;
}

######################################################################
# Modify hrEduOrg attributes (o, l, postalAddress) in all user entries
######################################################################
sub _ldapOrgModifyAll {
    my ($ldap, $base, $attribute, $values, $type) = @_;
    my ($log, $filter, $mesg, $code, $error, @entries, $out_msg, $dn, $num);
    my ($code_sum, $ent, $mesg2, $code2, $error2);
    $log      = "_ldapOrgModifyAll";
    $out_msg  = '';
    $code_sum = 0;

    AOSICFG::write_ws_log("/ $log(in):attribute=$attribute;values=" . AOSICFG::dump_array($values)) if $config::aosi_debug_level > 2;
    $filter = '(uid=*)';
    $mesg   = $ldap->search(base => $base, filter => $filter, timelimit => 60);
    $code   = $mesg->code;
    $error  = $mesg->error;
    if ($code == 0) {
        @entries = $mesg->entries;
        foreach $ent (@entries) {
            $dn = $ent->dn();
            AOSICFG::write_ws_log("| $log: changing for DN: $dn;") if $config::aosi_debug_level > 3;
            if ($type == 1) {    # Add
                $mesg2 = $ldap->modify($dn, add => { $attribute => $values });
            }
            elsif ($type == 2) {    # Replace
                $mesg2 = $ldap->modify($dn, replace => { $attribute => $values });
            }
            elsif ($type == 3) {    # Delete
                $mesg2 = $ldap->modify($dn, delete => { $attribute => $values });
            }
            $code2  = $mesg2->code;
            $error2 = $mesg2->error;
            if ($code2 != 0) {
                $out_msg .= 'Failed for DN ' . $dn . '!';
                $code_sum += $code2;
                AOSICFG::write_ws_log("| $log: Failed for DN: $dn (code=$code2;error=$error2);") if $config::aosi_debug_level > 3;
            }
        }
    }
    else {
        $out_msg  = 'User change error: ' . $code . ': ' . $error;
        $code_sum = $code;
    }

    AOSICFG::write_ws_log("\\ $log(out): code_sum=$code_sum;out_msg=$out_msg;") if $config::aosi_debug_level > 2;
    return $code_sum, $out_msg;
}

######################################################################
# Modify hrEduOrg attributes
######################################################################
sub ldapOrgGenModify {
    my ($service, $user, $pwd, $base, $xml_in, $type) = @_;
    my ($hash, $log, $filter, $attributes, $auth_code, $auth_error, $is_admin, $admin_filter);
    my ($entries, $attr_exists, @entries, $dn_in, $out_one, $out_msg, $attrs, $len, $att, $val, $mesg, $code, $error);
    my ($hostname, $port, $ldap, $ldap_error, $mesg2, $ent, $i, $j, @attrs, @vals, $mod_hash);
    my ($msg2, $code2, $org_att, $all_msg, $all_code, @all_mod, $has_all_org, $old_hash, $old_xml, $dummy, $old_ent, $old_val);
    my (@new_vals, @old_vals, $cur_val, $mod_hash2, $code3, $msg3,$restriction_code, $restriction_msg, $real_val);
    my ($bin_hash, $bin_xml, $cur_code, $bin_ent, $bin_attr_ex, $bin_val, @hep_af_vals);
    $log         = "ldapOrgGenModify";
    $has_all_org = 0;

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;type:$type;xml:$xml_in;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                  if $config::aosi_debug_level > 1;

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = AOSICFG::ldap_init($hostname, $port);

    ($auth_code, $auth_error, $is_admin, $admin_filter) = AOSICFG::ck_auth($ldap, AOSICFG::get_dn($user, $base), $pwd, 1, $ldap_error);
    Encode::_utf8_on($admin_filter);
    if ($is_admin == 0 and $ldap) {
        if ($auth_code == 0) {
            $auth_code  = -15;
            $auth_error = "User $user does not have admin privileges!";
        }
    }
    if ($auth_code == 0) {
        $mesg2 = $ldap->bind($config::ldap_user, password => AOSICFG::get_pwd($config::ldap_pwd_file), version => 3);
        AOSICFG::write_ws_log("| $log(bind_as_admin): " . $mesg2->code) if $config::aosi_debug_level > 2;
        if ($mesg2->code != 0) {
            AOSICFG::write_ws_log("| !!!$log: invalid admin password in " . $config::ldap_pwd_file);
        }
        ($entries, $attr_exists) = AOSICFG::xml2entries($xml_in);
        @entries = @{$entries};

        $out_msg = '';
        foreach $ent (@entries) {
            $dn_in   = AOSICFG::utfize($ent->{'dn'});
            $out_one = '<entry dn="' . $dn_in . '">';
            $attrs   = $ent->{'attrs'};
            $len     = $#{$attrs};
            foreach $i (0 .. $len) {
                $att = $attrs->[$i]->{'att'};
                push(@attrs, $att);
                $val = $attrs->[$i]->{'val'};
                if (ref($val) eq 'ARRAY') {
                    @vals = @$val;
                    if ($#vals == -1) {
                        $vals[0] = '';
                    }
                    AOSICFG::write_ws_log("| $log($att; multi:" . $#vals . "): " . join(", ", @vals)) if $config::aosi_debug_level > 2;
                }
                else {
                    if ($val) {
                        push(@vals, $val);
                    }
                    AOSICFG::write_ws_log("| $log($att; single): $val") if $config::aosi_debug_level > 2;
                }
                $real_val = join(' ', @vals);
                    $code3 = 0;
                    if ($att eq 'facsimileTelephoneNumber') {
                        $old_hash = AOSI::ldapOrgInfo($log, $user, $pwd, $base, 1);
                        AOSICFG::write_ws_log("| $log(old_hash):" . AOSICFG::dump_hash($old_hash)) if $config::aosi_debug_level > 2;
                        $old_xml = $old_hash->{'result'};
                        ($old_ent, $dummy) = xml2entries($old_xml);
                        AOSICFG::write_ws_log("| $log: old_ent=" . join(",", @{ $old_ent->[0]->{'attrs'} })) if $config::aosi_debug_level > 2;
                        foreach $cur_val (@{ $old_ent->[0]->{'attrs'} }) {
                            if ($cur_val->{'att'} eq $att) {
                                $old_val = $cur_val->{'val'};
                            }
                        }
                        if ($old_val) {
                            @old_vals = @$old_val;
                        }
                        else {
                            @old_vals = ();
                        }
                        if ($type == 1) {    # Add
                            @new_vals = @old_vals;
                            AOSICFG::write_ws_log("| $log(fTN_add A): (" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                            foreach $cur_val (@vals) {
                                $dummy = 0;
                                foreach $old_val (@old_vals) {
                                    if ($cur_val eq $old_val) {
                                        $dummy++;
                                    }
                                }
                                if ($dummy == 0) {
                                    push(@new_vals, $cur_val);
                                }

                            }
                            AOSICFG::write_ws_log("| $log(fTN_add B): (" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                            foreach $cur_val (@{ $mod_hash2->{$att} }) {
                                push(@new_vals, $cur_val);
                            }
                            AOSICFG::write_ws_log("| $log(fTN_add C): (" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                        }
                        elsif ($type == 2) {    # Modify
                            @new_vals = @vals;
                            foreach $cur_val (@{ $mod_hash2->{$att} }) {
                                push(@new_vals, $cur_val);
                            }
                            AOSICFG::write_ws_log("| $log(fTN_mod): (" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                        }
                        elsif ($type == 3) {    # delete
                            foreach $old_val (@old_vals) {
                                foreach $val (@vals) {
                                    if ($old_val ne $val) {
                                        push(@new_vals, $old_val);
                                    }
                                }
                            }
                            if ($#old_vals > -1 and ($#old_vals != $#new_vals)) {
                                AOSICFG::write_ws_log("| $log(fTN_del): (" . join(",", @new_vals) . ")") if $config::aosi_debug_level > 2;
                            }
                            else {
                                AOSICFG::write_ws_log("| $log(fTN_del): nothing to delete.") if $config::aosi_debug_level > 2;
                            }
                        }
                        AOSICFG::write_ws_log("| $log(fTN_new_vals): " . join(", ", @new_vals)) if $config::aosi_debug_level > 2;
                        $mod_hash2->{$att} = \@new_vals;
                        AOSICFG::write_ws_log("| $log(mod_fTN):" . AOSICFG::dump_hash($mod_hash2)) if $config::aosi_debug_level > 2;
                        $mesg2 = $ldap->modify($dn_in, delete => [$att]);
                        AOSICFG::write_ws_log("| $log(mod_fTN del):code:" . $mesg2->code . ";error:" . $mesg2->error) if $config::aosi_debug_level > 3;
                        $mesg2 = $ldap->modify($dn_in, add => $mod_hash2);
                        AOSICFG::write_ws_log("| $log(mod_fTN replace):code:" . $mesg2->code . ";error:" . $mesg2->error) if $config::aosi_debug_level > 3;
                    }
                    elsif ($att eq 'hrEduOrgOIB') {
                        ($restriction_code, $restriction_msg) = is_restriction_ok($att, $real_val);
                        if ($restriction_code > 0) {
                            if($type == 2) {    # Replace
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(objectclass=organization)", 'hrEduOrgUniqueNumber', 0, -1, 0);
                                $bin_xml  = $bin_hash->{'result'};
                                $cur_code = $bin_hash->{'code'};
                                ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                if (ref($bin_val) eq 'ARRAY') {
                                    @hep_af_vals = @$bin_val;
                                }
                                else {
                                    @hep_af_vals = ($bin_val);
                                }
                                AOSICFG::write_ws_log("| $log(hep_UN get OIB): code:$cur_code;values:[" . join(",", @hep_af_vals) . "]")
                                  if $config::aosi_debug_level > 3;
                                if($#hep_af_vals > -1){   # There are existing values in hrEduOrgUniqueNumber
                                    foreach $j (0..$#hep_af_vals){
                                        $key = $hep_af_vals[$j];
                                        if($key =~ /^OIB:\s*(\w+)/){    # There is an OIB valeu in hrEduOrgUniqueNumber
                                            splice(@hep_af_vals, $j, 1);
                                        }
                                    }
                                }
                                push(@{ $mod_hash->{$att} }, $real_val);
                                push(@hep_af_vals, 'OIB: '.$real_val);   # Final array off all values for hrEduPersonUniqueNumber forwarded to plug-ins
                                push(@{ $mod_hash->{'hrEduOrgUniqueNumber'} }, @hep_af_vals);
                                 AOSICFG::write_ws_log("| $log(hep_UN final hrEduOrgUniqueNumber): values:[" . join(",", @hep_af_vals) . "]")
                                  if $config::aosi_debug_level > 2;
                                $code3 = 0;
                            }
                            else {    # Other (Add nad Delete)
                                push(@{ $mod_hash->{$att} }, $real_val);
                            }
                        }
                        else {
                            $code3 = -65;
                            $msg3 = $restriction_msg;
                        }
                        AOSICFG::write_ws_log("| $log(hrEduOrgOIB out): code:$code3;msg:$msg3")
                              if $config::aosi_debug_level > 2;
                    }
                    elsif ($att eq 'hrEduOrgUniqueNumber') {
                        if($type == 2){   # Replace
                            # Fetch any existing hrEduOrgUniqueNumber values for replace
                            $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(objectclass=organization)", 'hrEduOrgUniqueNumber', 0, -1, 0);
                            $bin_xml  = $bin_hash->{'result'};
                            $cur_code = $bin_hash->{'code'};
                            ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                            $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                            if (ref($bin_val) eq 'ARRAY') {
                                @hep_af_vals = @$bin_val;
                            }
                            else {
                                @hep_af_vals = ($bin_val);
                            }
                            AOSICFG::write_ws_log("| $log(hep_UN get hrEduOrgUniqueNumber): code:$cur_code;values:[" . join(",", @hep_af_vals) . "]")
                              if $config::aosi_debug_level > 3;
                            if($#hep_af_vals > -1){   # There are existing values in hrEduOrgUniqueNumber
                                foreach $key (@hep_af_vals){
                                    if($key =~ /^OIB:\s*(\w+)/){
                                        $bin_val = $key;    # Get OIB value from existing hrEduOrgUniqueNumber
                                    }
                                }
                            }
                            foreach $key (@vals){
                               if($key =~ /^OIB:\s*(\w+)/){
                                    $real_val = $key;       # Get OIB value from new values
                                }
                            }
                            AOSICFG::write_ws_log("| $log(prev vs cur): $bin_val vs $real_val") if $config::aosi_debug_level > 3;
                            if($real_val ne $bin_val){
                                $code3 = -66;
                                $msg3  = 'OIB in hrEduOrgUniqueNumber cannot be changed! Use hrEduOrgOIB to change the OIB value of hrEduOrgUniqueNumber.';
                            }
                            else {
                              push(@{ $mod_hash->{$att} }, @vals);
                            }
                        }
                        else {
                            push(@{ $mod_hash->{$att} }, @vals);
                        }
                    }
                    else {
                        push(@{ $mod_hash->{$att} }, @vals);
                    }
            }
            if($code3 == 0){
                AOSICFG::write_ws_log("| $log(mod_hash):" . AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                if ($type == 1) {
                    $mesg = $ldap->modify($dn_in, add => $mod_hash);
                }
                elsif ($type == 2) {
                    $mesg = $ldap->modify($dn_in, replace => $mod_hash);
                }
                elsif ($type == 3) {
                    $mesg = $ldap->modify($dn_in, delete => $mod_hash);
                }
                $code  = $mesg->code;
                $error = $mesg->error;
                if ($code == 0) {
                    foreach $org_att ('o', 'l', 'postalAddress') {
                        if ($mod_hash->{$org_att} ne '') {
                            $has_all_org++;
                            ($code2, $msg2) = _ldapOrgModifyAll($ldap, $base, $org_att, $mod_hash->{$org_att}, $type);
                            $all_code += $code2;
                            $all_msg .= $msg2;
                            push(@all_mod, $org_att);
                        }
                    }
                    if ($has_all_org) {
                        if ($all_code == 0) {
                            $all_msg = ' Modified ' . join(", ", @all_mod) . ' for all users!';
                        }
                        else {
                            $all_msg = ' Modified ' . join(", ", @all_mod) . ' for all users but: ' . $all_msg . '!';
                        }
                    }
                    $out_one .= '<status code="0">Modified attribute ' . join(", ", @attrs) . '!' . $all_msg . '</status>';
                }
                else {
                    $out_one .= '<status code="' . $code . '">' . $error . '!</status>';
                }
            }
            else {
                $code += abs($code3);
                $out_one .= '<status code="' . $code3 . '">' . $msg3 . '!</status>';
                $err_num++;
            }
            $out_one .= '</entry>';
            $out_msg .= $out_one;
        }

        $hash->{'code'}   = $code;
        $hash->{'error'}  = $error;
        $hash->{'result'} = '<?xml version="1.0"?><ldap action="' . $service . '" version="' . $config::aosi_version . '">' . $out_msg . '</ldap>';
        AOSICFG::write_action_log($user, $log, $config::base_dn, $code, join(", ", @attrs));
    }
    else {
        $hash->{'code'}  = $auth_code;
        $hash->{'error'} = $auth_error;
        $hash->{'result'} =
            '<?xml version="1.0"?><ldap action="' . $service
          . '" version="'
          . $config::aosi_version
          . '"><entry dn="'
          . $dn_in
          . '"><status code="'
          . $auth_code . '">'
          . $auth_error
          . '</status></entry></ldap>';
    }

    if ($type > 0) {
        AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
        return $hash;
    }
    else {
        AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
# Makes XML from hash
# Input:    hash reference
#           type string (must be before or after)
# Output:   panic flag value
#           plugin code value
#           XML
######################################################################
sub process_plugin_hash {
    my ($plugin_hash, $type) = @_;    # $type must be 'before' or 'after'
    my ($plugin_panic, $plugin_code, $plugin_xml, $temp);
    $plugin_panic = 0;
    $plugin_code  = 0;
    $plugin_xml   = '';

    if (($type eq 'before') or ($type eq 'after')) {

        # The $temp hack is necessary so $plugin_xml stays UTF-8
        foreach $plugin_key (keys %$plugin_hash) {
            $plugin_panic += $plugin_hash->{$plugin_key}->{$type}->{'panic'};
            $plugin_code  += abs($plugin_hash->{$plugin_key}->{$type}->{'code'});
            $temp = '<' . $type . '_plugin_status name="' . $plugin_key . '" code="' . $plugin_hash->{$plugin_key}->{$type}->{'code'} . '">';
            Encode::_utf8_on($temp);
            $plugin_xml .= $temp;
            $temp = $plugin_hash->{$plugin_key}->{$type}->{'msg'};
            Encode::_utf8_on($temp);
            $plugin_xml .= $temp;
            $temp = '</' . $type . '_plugin_status>';
            Encode::_utf8_on($temp);
            $plugin_xml .= $temp;
        }
    }
    elsif ($type eq 'search') {
        $plugin_panic = '';
        $plugin_code  = 0;
        $plugin_xml   = '';
        foreach $plugin_key (keys %$plugin_hash) {
            $plugin_code += abs($plugin_hash->{$plugin_key}->{$type}->{'code'});
            $temp = $plugin_hash->{$plugin_key}->{$type}->{'msg'};
            Encode::_utf8_on($temp);
            $plugin_panic .= $temp;
            $temp = $plugin_hash->{$plugin_key}->{$type}->{'xml'};
            Encode::_utf8_on($temp);
            $plugin_xml .= $temp;
        }
    }
    return $plugin_panic, $plugin_code, $plugin_xml;
}

######################################################################
# Encodes input string in SHA algorithm if it is not encoded allready
# Input:    string to encode
# Output:   encoded string
######################################################################
sub encode_pwd {
    my ($in) = @_;
    my ($out, $log);
    $log = (caller(0))[3];

    #AOSICFG::write_ws_log("| $log(in): $in;") if $config::aosi_debug_level > 3;
    if ($in =~ /^\{SHA\}.*/) {
        $out = $in;
    }
    else {
        $out = '{SHA}' . sha1_base64(encode_utf8($in)) . '=';
    }
    AOSICFG::write_ws_log("| $log(out): $out;") if $config::aosi_debug_level > 3;
    return $out;
}

######################################################################
#
######################################################################
sub ck_auth {
    my ($ldap, $dn, $pwd, $ck_admin, $ldap_error) = @_;
    my ($adm, $admin, $admin_type, $attrib, $base, $code, $dc, $dec_pwd);
    my ($domain,     $error, $full_uid, $hredu_uid, $is_admin_ldap, $mesg);
    my ($rest,       $tld,   $type,     $user,      %options,       @entries, $entry, $log);
    my ($out_filter, $val,   $dummy,    @urn,       $urn_val,       $urn_key, @or, @keys, %and, $temp);
    my (@entries2, $entry2, @values2, $filter, $ppolicy, $pp_resp, $pp_mesg);
    $log = 'ck_auth';

    $is_admin_ldap = 0;
    $out_filter    = '';
    AOSICFG::write_ws_log("|/ $log(in): dn=$dn;ck_admin=$ck_admin;ldap=$ldap;ip=$AOSI::client_ip;direct_access:" . $config::direct_access);
    if ($ldap) {
        if ($dn eq '' || $pwd eq '') {
            AOSICFG::write_ws_log("|| $log: No DN or password") if $config::aosi_debug_level > 3;
            $type = 'IP';
            if ($config::direct_access == 1) {
                AOSICFG::write_ws_log("|| $log: using direct access") if $config::aosi_debug_level > 3;
                if (AOSICFG::ip_is_valid($AOSI::client_ip)) {
                    AOSICFG::write_ws_log("|| $log: using direct access OK!") if $config::aosi_debug_level > 3;
                    $is_admin_ldap = 1;
                    $code          = 0;
                }
                else {
                    if($pwd eq ''){
                        $code  = -10;
                        $error = 'No password (using direct access) from IP: ' . $AOSI::client_ip . '!';
                        AOSICFG::write_ws_log("|| $log(no_pwd):using valid ip=INVALID") if $config::aosi_debug_level > 3;
                    }
                    else {
                        $code  = -14;
                        $error = 'No DN specified (using direct access)!';
                        AOSICFG::write_ws_log("|| $log(no_dn):") if $config::aosi_debug_level > 3;
                    }
                }
            }
            else {
                if ($dn eq '') {
                    $code  = -12;
                    $error = 'No DN specified!';
                    AOSICFG::write_ws_log("|| $log(no_dn_pwd): No DN") if $config::aosi_debug_level > 3;
                }
                elsif ($pwd eq '') {
                    $code  = -13;
                    $error = 'No password specified!';
                    AOSICFG::write_ws_log("|| $log(no_dn_pwd): No password") if $config::aosi_debug_level > 3;
                }
            }
        }
        else {
            $type    = 'user/pwd';
            $dec_pwd = decode_base64($pwd);
			$ppolicy = Net::LDAP::Control::PasswordPolicy->new;
			$mesg    = $ldap->bind($dn, password => $dec_pwd, version => 3, control => [ $ppolicy ]);
            $code    = $mesg->code;
            $error   = $mesg->error;
			$pp_mesg = '';
			$pp_resp = $mesg->control( LDAP_CONTROL_PASSWORDPOLICY );
			if(defined($pp_resp)){
				if(defined($pp_resp->pp_error)){
					$pp_mesg = AOSICFG::ppolicy_get_description($pp_resp->pp_error);
				}
				if(defined($pp_resp->time_before_expiration)){
					$pp_mesg = "Lozinka istjece za ".AOSICFG::convert_time($pp_resp->time_before_expiration).".";
				}
				if(defined($pp_resp->grace_authentications_remaining)){
					$pp_mesg = "Broj autentikacija nakon kojih ce se racun zakljucati: ". $pp_resp->grace_authentications_remaining."."
				}
			}
            AOSICFG::write_ws_log("|| $log(step one - bind):$dn;code:$code;error:$error") if $config::aosi_debug_level > 2;
            if ($code == 0) {    # Second auth step
                $attrib = 'hrEduPersonUniqueID';    #'hrEduPersonUniqueID'
                ($user, $domain, $tld) = split(/,/, $dn);
                $base   = "$domain,$tld";
                if($user =~ /^cn=/g){     # This is not a person
                    AOSICFG::write_ws_log("|| $log(step two - cn):dn=$dn;user=$user;") if $config::aosi_debug_level > 2;
                    $code = 0;
                }
                else {
                    ($rest, $user) = split(/=/, $user);
                    $filter = "(uid=$user)";
                    ($dc, $domain) = split(/=/, $domain);
                    ($dc, $tld)    = split(/=/, $tld);
                    $full_uid = $user . '@' . $domain . '.' . $tld;
                    $full_uid = lc($full_uid);
                    %options  = (
                        base   => $base,
                        filter => $filter,

                        #sizelimit => 0,
                        timelimit => 180,
                        attrs     => [$attrib]
                    );
                    $mesg      = $ldap->search(%options);
                    $code      = $mesg->code;
                    $error     = $mesg->error;
                    @entries   = $mesg->entries;
                    $hredu_uid = $entries[0]->get_value($attrib);
                    $hredu_uid = lc($hredu_uid);

                    #($hredu_uid, $rest) = split(/@/, $hredu_uid);
                    AOSICFG::write_ws_log("|| $log(step two - ck_uid):$hredu_uid ? $full_uid;code:$code;error:$error") if $config::aosi_debug_level > 2;
                    if ($hredu_uid eq $full_uid) {
                        $code = 0;
                    }
                    else {
                        $code  = -11;
                        $error = 'hrEduPersonUniqueID($hredu_uid) does not match $full_uid!';
                    }
                }
            }
            if ($code == 0 and $ck_admin == 1) {
                $out_filter = '';
                ($user, $rest) = split(/,/, $dn);
                ($rest, $user) = split(/=/, $user);
                $admin = AOSICFG::user_is_valid($user);
                if ($admin > 0) {
                    $is_admin_ldap = 1;
                    $admin_type    = "file";
                    AOSICFG::write_ws_log("|| $log(admin check - file): OK") if $config::aosi_debug_level > 2;
                }
                else {
                    $attrib  = 'hrEduPersonRole';
                    %options = (
                        base   => $base,
                        filter => "(uid=$user)",
                        timelimit => 60,
                        attrs     => [$attrib]
                    );
                    $mesg  = $ldap->search(%options);
                    $code  = $mesg->code;
                    $error = $mesg->error;
                    AOSICFG::write_ws_log("|| $log(search for hrEduPersonRole 1):uid=$user;base=$base;code:$code;error:$error") if $config::aosi_debug_level > 2;
                    @entries = $mesg->entries;
                    AOSICFG::write_ws_log("|| $log(search for hrEduPersonRole 2):entries:" . $#entries) if $config::aosi_debug_level > 2;
                    if($#entries > -1){
                        $entry = $entries[0];
                        @values = @{ $entry->get_value($attrib, asref => 1) };
                      LOOP:

                        foreach $val (@values) {
                            AOSICFG::write_ws_log("|| $log(hrEduPersonRole val):$val") if $config::aosi_debug_level > 3;
                            my %and;
                            if ($val =~ /^urn\:*/) {
                                AOSICFG::write_ws_log("|| $log(hrEduPersonRole val): has urn") if $config::aosi_debug_level > 3;
                                $is_admin_ldap = 3;
                                $admin_type    = "hrEduPersonRole=urn";
                                @urn           = split(/:/, $val);
                                $dummy         = shift(@urn);
                                $dummy         = shift(@urn);
                                if ($#urn > -1) {
                                    while ($urn_key = shift(@urn)) {
                                        $urn_val = shift(@urn);
                                        if ($urn_val ne '') {
                                            push(@{ $and{$urn_key} }, $urn_val);
                                        }
                                    }
                                }
                                else {
                                    @or = ();
                                    last LOOP;
                                }
                                push(@or, \%and);
                            }
                            if ($val eq 'administrator imenika') {
                                AOSICFG::write_ws_log("|| $log(hrEduPersonRole val): has urn") if $config::aosi_debug_level > 3;
                                $is_admin_ldap = 2;
                                $admin_type    = "hrEduPersonRole";
                                @or            = ();
                                last LOOP;
                            }
                        }
                        if ($#or > -1) {
                            foreach $val (@or) {
                                $dummy = '';
                                @keys  = keys %$val;
                                foreach $urn_key (@keys) {
                                    @urn = @{ $val->{$urn_key} };
                                    if ($#urn > 0) {
                                        $dummy .= '(|';
                                        foreach $urn_val (@urn) {
                                            $dummy .= '(' . $urn_key . '=' . $urn_val . ')';
                                        }
                                        $dummy .= ')';
                                    }
                                    else {
                                        $dummy .= '(' . $urn_key . '=' . $urn[0] . ')';
                                    }
                                }
                                if ($dummy ne '') {
                                    $out_filter .= '(&' . $dummy . ')';
                                }
                            }
                            $out_filter = '(|' . $out_filter . ')';
                        }
                    }
                    else {
                        $is_admin_ldap = 0;
                    }

                    if ($is_admin_ldap == 0) {
                        $admin_type = "none";
                    }
                    AOSICFG::write_ws_log("|| $log(admin check - $attrib): OK") if $config::aosi_debug_level > 2;
                }
                $type .= " (admin: $admin_type)";
            }
        }
    }
    else {
        $code  = -18;
        $error = $ldap_error;
    }
    AOSICFG::write_ws_log("|\\ $log(out):type:$type;code:$code;error:$error;out_filter:$out_filter");

    return $code, $error, $is_admin_ldap, $out_filter, $pp_mesg, \@or;
}

sub ck_jmbg {
    my($jmbg) = @_;
    my($out, $len, $last1, $rest, $last2);

    $last1 = substr($jmbg,-1);
    $len = length($jmbg);
    if($len == 13){
        $rest = substr($jmbg, 0, $len-1);
        $last2 = jmbg_ctrl_num($rest);
        AOSICFG::write_ws_log("|| ck_jmbg: jmgb=$jmbg; last1=$last1; last2=$last2;") if $config::aosi_debug_level > 3;
        $out = $last1 eq $last2 ? 1 : 0;
    }
    else {
      $out = 0;
    }

    return $out;
}

sub ck_oib {
    my($oib) = @_;
    my($out, $len, $last1, $rest, $last2);

    $last1 = substr($oib,-1);
    $len = length($oib);
    if($len == 11){
        $rest = substr($oib, 0, $len-1);
        $last2 = mod1110_ctrl_num($rest);
        AOSICFG::write_ws_log("|| ck_oib: oib=$oib; last1=$last1; last2=$last2;") if $config::aosi_debug_level > 3;
        $out = $last1 eq $last2 ? 1 : 0;
    }
    else {
      $out = 0;
    }

    return $out;
}

sub jmbg_ctrl_num {
    my($in_stg) = @_;
    my($num, $len, $sum, $i, $rest, $out);

    $num = $in_stg;
    $num =~ s/\D//g;

    $len = length($num);


    $sum = 0;
    for ($i = 0; $i < $len; $i++) {
        $cur = substr($num, $i, 1);
        $tmp1 = $i % 6;
        $tmp2 = 7 - $tmp1;
        $tmp3 = $tmp2 * $cur;
        $sum += $tmp3;
    }
    $rest = $sum % 11;
    $out = $rest == 0 ? 0 : 11 - $rest;

    return $out;
}

sub mod1110_ctrl_num {
    my($in_stg, $in_mod1, $in_mod2) = @_;
    my($mod1, $mod2, $num, $len, $a, $i, $out);

    $mod1 = $in_mod1 eq '' ? 11 : $in_mod1;
    $mod2 = $in_mod2 eq '' ? 10 : $in_mod2;
    $num = $in_stg;
    $num =~ s/\D//g;

    $len = length($num);

    $a = $mod2;
    for ($i = 0; $i < $len; $i++) {
    	$a = $a + substr($num, $i, 1);
    	$a = $a % $mod2;
    	if ( $a == 0 ) { $a = $mod2; }
    	$a *= 2;
    	$a = $a % $mod1;
    }
    $out = $mod1 - $a;
    $out = $out == $mod2 ? 0 : $out;

    return $out;
}

sub aosi_version {
    my ($type) = @_;
    my ($ver, $major, $minor, $date, $out, $day, $mon, $year);

    $type = defined $type ? $type : 1;

    $ver = $config::aosi_version;
    ($major, $minor, $date) = split(/\./, $ver);
    $year = substr($date, 0, 4);
    $mon = substr($date, 5, 2);
    $day = substr($date, 7, 2);

    if($type == 1){
        $out = $major.'.'.$minor;
    }
    elsif($type == 2){
        $out = $major;
    }
    elsif($type == 3){
        $out = $minor;
    }
    elsif($type == 4){
        $out = $date;
    }
    elsif($type == 5){
        $out = $year;
    }
    elsif($type == 6){
        $out = $mon;
    }
    elsif($type == 7){
        $out = $day;
    }

    AOSICFG::write_ws_log("|| aosi_version: type=$type;version=".$config::aosi_version.";out=$out;") if $config::aosi_debug_level > 3;
    return $out;
}

sub to_ascii {
  my ($in) = @_;
  my (%trans, $out, $key);

  AOSICFG::write_ws_log("|| to_ascii: in=$in") if $config::aosi_debug_level > 3;
  $out = $in;

  %trans = (
      # Croatian letters
      '0106' => 'C', # TJ
      '0107' => 'c', # tj
      '010C' => 'C', # CH
      '010D' => 'c', # ch
      '0110' => 'Dj', # DJ
      '0111' => 'dj', # dj
      '0160' => 'S', # SH
      '0161' => 's', # sh
      '017D' => 'Z', # ZH
      '017E' => 'z', # zh
      # German letters
      '00C4' => 'Ae',  # AE
      '00E4' => 'ae',  # ae
      '00D6' => 'Oe',  # OE
      '00F6' => 'oe',  # oe
      '00DC' => 'Ue',  # UE
      '00FC' => 'ue',  # ue
      '00DF' => 'ss',  # ss
  );
  foreach $key (keys %trans){
      $out =~ s/\x{$key}/$trans{$key}/g;
  }
  AOSICFG::write_ws_log("|| to_ascii: out=$out") if $config::aosi_debug_level > 3;
  return $out;
}

sub ppolicy_get_description {
    my($id, $raw) = @_;
    my($pp_const, $out);
    
    $pp_const = {
        0 => 'Loznika je istekla',    #'LDAP_PP_PASSWORD_EXPIRED',
        1 => 'Racun je zakljucan',    #'LDAP_PP_ACCOUNT_LOCKED',
        2 => 'Izmjena loznike nakon postavljanja', #'LDAP_PP_CHANGE_AFTER_RESET',
        3 => 'Zabranjena izmjena lozinke', #'LDAP_PP_PASSWORD_MOD_NOT_ALLOWED',
        4 => 'Nedostaje stara lozinka',   #'LDAP_PP_MUST_SUPPLY_OLD_PASSWORD',
        5 => 'Lozinka nije dovoljno kvalitetna', #'LDAP_PP_INSUFFICIENT_PASSWORD_QUALITY',
        6 => 'Loznika je prekratka', #'LDAP_PP_PASSWORD_TOO_SHORT',
        7 => 'Lozinka je premlada',  #'LDAP_PP_PASSWORD_TOO_YOUNG',
        8 => 'Lozinka je iz niza prethodnih', #'LDAP_PP_PASSWORD_IN_HISTORY'
    };
        
    $out = $pp_const->{$id};
    if( ! defined($raw)){
        $out =~ s/^LDAP_PP_//;        
    }
    return $out;    
}

sub convert_time {
    my $time = shift;
    my $days = int($time / 86400);
    $time -= ($days * 86400);
    my $hours = int($time / 3600);
    $time -= ($hours * 3600);
    my $minutes = int($time / 60);
    my $seconds = $time % 60;

    $days = $days < 1 ? '' : $days .'d ';
    $hours = $hours < 1 ? '' : $hours .'h ';
    $minutes = $minutes < 1 ? '' : $minutes . 'm ';
    $time = $days . $hours . $minutes . $seconds . 's';
    return $time;
}
1;
